
import google.protobuf
from packaging.version import Version
if Version(google.protobuf.__version__).major >= 5:
  # -*- coding: utf-8 -*-
  # Generated by the protocol buffer compiler.  DO NOT EDIT!
  # source: facet_feature_statistics.proto
  # Protobuf Python Version: 5.26.0
  """Generated protocol buffer code."""
  from google.protobuf import descriptor as _descriptor
  from google.protobuf import descriptor_pool as _descriptor_pool
  from google.protobuf import symbol_database as _symbol_database
  from google.protobuf.internal import builder as _builder
  # @@protoc_insertion_point(imports)

  _sym_db = _symbol_database.Default()




  DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1e\x66\x61\x63\x65t_feature_statistics.proto\x12\x16\x66\x61\x63\x65tFeatureStatistics\"b\n\x1c\x44\x61tasetFeatureStatisticsList\x12\x42\n\x08\x64\x61tasets\x18\x01 \x03(\x0b\x32\x30.facetFeatureStatistics.DatasetFeatureStatistics\"\x14\n\x04Path\x12\x0c\n\x04step\x18\x01 \x03(\t\"\x9e\x01\n\x18\x44\x61tasetFeatureStatistics\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0cnum_examples\x18\x02 \x01(\x03\x12\x1d\n\x15weighted_num_examples\x18\x04 \x01(\x01\x12?\n\x08\x66\x65\x61tures\x18\x03 \x03(\x0b\x32-.facetFeatureStatistics.FeatureNameStatistics\"\xae\x04\n\x15\x46\x65\x61tureNameStatistics\x12\x0e\n\x04name\x18\x01 \x01(\tH\x00\x12,\n\x04path\x18\x08 \x01(\x0b\x32\x1c.facetFeatureStatistics.PathH\x00\x12@\n\x04type\x18\x02 \x01(\x0e\x32\x32.facetFeatureStatistics.FeatureNameStatistics.Type\x12>\n\tnum_stats\x18\x03 \x01(\x0b\x32).facetFeatureStatistics.NumericStatisticsH\x01\x12@\n\x0cstring_stats\x18\x04 \x01(\x0b\x32(.facetFeatureStatistics.StringStatisticsH\x01\x12>\n\x0b\x62ytes_stats\x18\x05 \x01(\x0b\x32\'.facetFeatureStatistics.BytesStatisticsH\x01\x12@\n\x0cstruct_stats\x18\x07 \x01(\x0b\x32(.facetFeatureStatistics.StructStatisticsH\x01\x12=\n\x0c\x63ustom_stats\x18\x06 \x03(\x0b\x32\'.facetFeatureStatistics.CustomStatistic\"=\n\x04Type\x12\x07\n\x03INT\x10\x00\x12\t\n\x05\x46LOAT\x10\x01\x12\n\n\x06STRING\x10\x02\x12\t\n\x05\x42YTES\x10\x03\x12\n\n\x06STRUCT\x10\x04\x42\n\n\x08\x66ield_idB\x07\n\x05stats\"x\n\x18WeightedCommonStatistics\x12\x17\n\x0fnum_non_missing\x18\x01 \x01(\x01\x12\x13\n\x0bnum_missing\x18\x02 \x01(\x01\x12\x16\n\x0e\x61vg_num_values\x18\x03 \x01(\x01\x12\x16\n\x0etot_num_values\x18\x04 \x01(\x01\"\xc3\x01\n\x0f\x43ustomStatistic\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x03num\x18\x02 \x01(\x01H\x00\x12\r\n\x03str\x18\x03 \x01(\tH\x00\x12\x36\n\thistogram\x18\x04 \x01(\x0b\x32!.facetFeatureStatistics.HistogramH\x00\x12?\n\x0erank_histogram\x18\x05 \x01(\x0b\x32%.facetFeatureStatistics.RankHistogramH\x00\x42\x05\n\x03valJ\x04\x08\x06\x10\x07\"\xb9\x02\n\x11NumericStatistics\x12>\n\x0c\x63ommon_stats\x18\x01 \x01(\x0b\x32(.facetFeatureStatistics.CommonStatistics\x12\x0c\n\x04mean\x18\x02 \x01(\x01\x12\x0f\n\x07std_dev\x18\x03 \x01(\x01\x12\x11\n\tnum_zeros\x18\x04 \x01(\x03\x12\x0b\n\x03min\x18\x05 \x01(\x01\x12\x0e\n\x06median\x18\x06 \x01(\x01\x12\x0b\n\x03max\x18\x07 \x01(\x01\x12\x35\n\nhistograms\x18\x08 \x03(\x0b\x32!.facetFeatureStatistics.Histogram\x12Q\n\x16weighted_numeric_stats\x18\t \x01(\x0b\x32\x31.facetFeatureStatistics.WeightedNumericStatistics\"\xa0\x03\n\x10StringStatistics\x12>\n\x0c\x63ommon_stats\x18\x01 \x01(\x0b\x32(.facetFeatureStatistics.CommonStatistics\x12\x0e\n\x06unique\x18\x02 \x01(\x03\x12I\n\ntop_values\x18\x03 \x03(\x0b\x32\x35.facetFeatureStatistics.StringStatistics.FreqAndValue\x12\x12\n\navg_length\x18\x04 \x01(\x02\x12=\n\x0erank_histogram\x18\x05 \x01(\x0b\x32%.facetFeatureStatistics.RankHistogram\x12O\n\x15weighted_string_stats\x18\x06 \x01(\x0b\x32\x30.facetFeatureStatistics.WeightedStringStatistics\x1aM\n\x0c\x46reqAndValue\x12\x1b\n\x0f\x64\x65precated_freq\x18\x01 \x01(\x03\x42\x02\x18\x01\x12\r\n\x05value\x18\x02 \x01(\t\x12\x11\n\tfrequency\x18\x03 \x01(\x01\"\x81\x01\n\x19WeightedNumericStatistics\x12\x0c\n\x04mean\x18\x01 \x01(\x01\x12\x0f\n\x07std_dev\x18\x02 \x01(\x01\x12\x0e\n\x06median\x18\x03 \x01(\x01\x12\x35\n\nhistograms\x18\x04 \x03(\x0b\x32!.facetFeatureStatistics.Histogram\"\xa4\x01\n\x18WeightedStringStatistics\x12I\n\ntop_values\x18\x01 \x03(\x0b\x32\x35.facetFeatureStatistics.StringStatistics.FreqAndValue\x12=\n\x0erank_histogram\x18\x02 \x01(\x0b\x32%.facetFeatureStatistics.RankHistogram\"\xac\x01\n\x0f\x42ytesStatistics\x12>\n\x0c\x63ommon_stats\x18\x01 \x01(\x0b\x32(.facetFeatureStatistics.CommonStatistics\x12\x0e\n\x06unique\x18\x02 \x01(\x03\x12\x15\n\ravg_num_bytes\x18\x03 \x01(\x02\x12\x15\n\rmin_num_bytes\x18\x04 \x01(\x02\x12\x15\n\rmax_num_bytes\x18\x05 \x01(\x02J\x04\x08\x06\x10\x07\"R\n\x10StructStatistics\x12>\n\x0c\x63ommon_stats\x18\x01 \x01(\x0b\x32(.facetFeatureStatistics.CommonStatistics\"\x88\x03\n\x10\x43ommonStatistics\x12\x17\n\x0fnum_non_missing\x18\x01 \x01(\x03\x12\x13\n\x0bnum_missing\x18\x02 \x01(\x03\x12\x16\n\x0emin_num_values\x18\x03 \x01(\x03\x12\x16\n\x0emax_num_values\x18\x04 \x01(\x03\x12\x16\n\x0e\x61vg_num_values\x18\x05 \x01(\x02\x12\x16\n\x0etot_num_values\x18\x08 \x01(\x03\x12?\n\x14num_values_histogram\x18\x06 \x01(\x0b\x32!.facetFeatureStatistics.Histogram\x12O\n\x15weighted_common_stats\x18\x07 \x01(\x0b\x32\x30.facetFeatureStatistics.WeightedCommonStatistics\x12H\n\x1d\x66\x65\x61ture_list_length_histogram\x18\t \x01(\x0b\x32!.facetFeatureStatistics.HistogramJ\x04\x08\n\x10\x0bJ\x04\x08\x0b\x10\x0c\"\xce\x02\n\tHistogram\x12\x0f\n\x07num_nan\x18\x01 \x01(\x03\x12\x15\n\rnum_undefined\x18\x02 \x01(\x03\x12\x39\n\x07\x62uckets\x18\x03 \x03(\x0b\x32(.facetFeatureStatistics.Histogram.Bucket\x12=\n\x04type\x18\x04 \x01(\x0e\x32/.facetFeatureStatistics.Histogram.HistogramType\x12\x0c\n\x04name\x18\x05 \x01(\t\x1a\x63\n\x06\x42ucket\x12\x11\n\tlow_value\x18\x01 \x01(\x01\x12\x12\n\nhigh_value\x18\x02 \x01(\x01\x12\x1c\n\x10\x64\x65precated_count\x18\x03 \x01(\x03\x42\x02\x18\x01\x12\x14\n\x0csample_count\x18\x04 \x01(\x01\",\n\rHistogramType\x12\x0c\n\x08STANDARD\x10\x00\x12\r\n\tQUANTILES\x10\x01\"\xce\x01\n\rRankHistogram\x12=\n\x07\x62uckets\x18\x01 \x03(\x0b\x32,.facetFeatureStatistics.RankHistogram.Bucket\x12\x0c\n\x04name\x18\x02 \x01(\t\x1ap\n\x06\x42ucket\x12\x10\n\x08low_rank\x18\x01 \x01(\x03\x12\x11\n\thigh_rank\x18\x02 \x01(\x03\x12\x1c\n\x10\x64\x65precated_count\x18\x03 \x01(\x03\x42\x02\x18\x01\x12\r\n\x05label\x18\x04 \x01(\t\x12\x14\n\x0csample_count\x18\x05 \x01(\x01')

  _globals = globals()
  _builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
  _builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'facet_feature_statistics_pb2', _globals)
  if not _descriptor._USE_C_DESCRIPTORS:
    DESCRIPTOR._loaded_options = None
    _globals['_STRINGSTATISTICS_FREQANDVALUE'].fields_by_name['deprecated_freq']._loaded_options = None
    _globals['_STRINGSTATISTICS_FREQANDVALUE'].fields_by_name['deprecated_freq']._serialized_options = b'\030\001'
    _globals['_HISTOGRAM_BUCKET'].fields_by_name['deprecated_count']._loaded_options = None
    _globals['_HISTOGRAM_BUCKET'].fields_by_name['deprecated_count']._serialized_options = b'\030\001'
    _globals['_RANKHISTOGRAM_BUCKET'].fields_by_name['deprecated_count']._loaded_options = None
    _globals['_RANKHISTOGRAM_BUCKET'].fields_by_name['deprecated_count']._serialized_options = b'\030\001'
    _globals['_DATASETFEATURESTATISTICSLIST']._serialized_start=58
    _globals['_DATASETFEATURESTATISTICSLIST']._serialized_end=156
    _globals['_PATH']._serialized_start=158
    _globals['_PATH']._serialized_end=178
    _globals['_DATASETFEATURESTATISTICS']._serialized_start=181
    _globals['_DATASETFEATURESTATISTICS']._serialized_end=339
    _globals['_FEATURENAMESTATISTICS']._serialized_start=342
    _globals['_FEATURENAMESTATISTICS']._serialized_end=900
    _globals['_FEATURENAMESTATISTICS_TYPE']._serialized_start=818
    _globals['_FEATURENAMESTATISTICS_TYPE']._serialized_end=879
    _globals['_WEIGHTEDCOMMONSTATISTICS']._serialized_start=902
    _globals['_WEIGHTEDCOMMONSTATISTICS']._serialized_end=1022
    _globals['_CUSTOMSTATISTIC']._serialized_start=1025
    _globals['_CUSTOMSTATISTIC']._serialized_end=1220
    _globals['_NUMERICSTATISTICS']._serialized_start=1223
    _globals['_NUMERICSTATISTICS']._serialized_end=1536
    _globals['_STRINGSTATISTICS']._serialized_start=1539
    _globals['_STRINGSTATISTICS']._serialized_end=1955
    _globals['_STRINGSTATISTICS_FREQANDVALUE']._serialized_start=1878
    _globals['_STRINGSTATISTICS_FREQANDVALUE']._serialized_end=1955
    _globals['_WEIGHTEDNUMERICSTATISTICS']._serialized_start=1958
    _globals['_WEIGHTEDNUMERICSTATISTICS']._serialized_end=2087
    _globals['_WEIGHTEDSTRINGSTATISTICS']._serialized_start=2090
    _globals['_WEIGHTEDSTRINGSTATISTICS']._serialized_end=2254
    _globals['_BYTESSTATISTICS']._serialized_start=2257
    _globals['_BYTESSTATISTICS']._serialized_end=2429
    _globals['_STRUCTSTATISTICS']._serialized_start=2431
    _globals['_STRUCTSTATISTICS']._serialized_end=2513
    _globals['_COMMONSTATISTICS']._serialized_start=2516
    _globals['_COMMONSTATISTICS']._serialized_end=2908
    _globals['_HISTOGRAM']._serialized_start=2911
    _globals['_HISTOGRAM']._serialized_end=3245
    _globals['_HISTOGRAM_BUCKET']._serialized_start=3100
    _globals['_HISTOGRAM_BUCKET']._serialized_end=3199
    _globals['_HISTOGRAM_HISTOGRAMTYPE']._serialized_start=3201
    _globals['_HISTOGRAM_HISTOGRAMTYPE']._serialized_end=3245
    _globals['_RANKHISTOGRAM']._serialized_start=3248
    _globals['_RANKHISTOGRAM']._serialized_end=3454
    _globals['_RANKHISTOGRAM_BUCKET']._serialized_start=3342
    _globals['_RANKHISTOGRAM_BUCKET']._serialized_end=3454
  # @@protoc_insertion_point(module_scope)

else:
  # -*- coding: utf-8 -*-
  # Generated by the protocol buffer compiler.  DO NOT EDIT!
  # source: facet_feature_statistics.proto
  """Generated protocol buffer code."""
  from google.protobuf import descriptor as _descriptor
  from google.protobuf import descriptor_pool as _descriptor_pool
  from google.protobuf import message as _message
  from google.protobuf import reflection as _reflection
  from google.protobuf import symbol_database as _symbol_database
  # @@protoc_insertion_point(imports)

  _sym_db = _symbol_database.Default()




  DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1e\x66\x61\x63\x65t_feature_statistics.proto\x12\x16\x66\x61\x63\x65tFeatureStatistics\"b\n\x1c\x44\x61tasetFeatureStatisticsList\x12\x42\n\x08\x64\x61tasets\x18\x01 \x03(\x0b\x32\x30.facetFeatureStatistics.DatasetFeatureStatistics\"\x14\n\x04Path\x12\x0c\n\x04step\x18\x01 \x03(\t\"\x9e\x01\n\x18\x44\x61tasetFeatureStatistics\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0cnum_examples\x18\x02 \x01(\x03\x12\x1d\n\x15weighted_num_examples\x18\x04 \x01(\x01\x12?\n\x08\x66\x65\x61tures\x18\x03 \x03(\x0b\x32-.facetFeatureStatistics.FeatureNameStatistics\"\xae\x04\n\x15\x46\x65\x61tureNameStatistics\x12\x0e\n\x04name\x18\x01 \x01(\tH\x00\x12,\n\x04path\x18\x08 \x01(\x0b\x32\x1c.facetFeatureStatistics.PathH\x00\x12@\n\x04type\x18\x02 \x01(\x0e\x32\x32.facetFeatureStatistics.FeatureNameStatistics.Type\x12>\n\tnum_stats\x18\x03 \x01(\x0b\x32).facetFeatureStatistics.NumericStatisticsH\x01\x12@\n\x0cstring_stats\x18\x04 \x01(\x0b\x32(.facetFeatureStatistics.StringStatisticsH\x01\x12>\n\x0b\x62ytes_stats\x18\x05 \x01(\x0b\x32\'.facetFeatureStatistics.BytesStatisticsH\x01\x12@\n\x0cstruct_stats\x18\x07 \x01(\x0b\x32(.facetFeatureStatistics.StructStatisticsH\x01\x12=\n\x0c\x63ustom_stats\x18\x06 \x03(\x0b\x32\'.facetFeatureStatistics.CustomStatistic\"=\n\x04Type\x12\x07\n\x03INT\x10\x00\x12\t\n\x05\x46LOAT\x10\x01\x12\n\n\x06STRING\x10\x02\x12\t\n\x05\x42YTES\x10\x03\x12\n\n\x06STRUCT\x10\x04\x42\n\n\x08\x66ield_idB\x07\n\x05stats\"x\n\x18WeightedCommonStatistics\x12\x17\n\x0fnum_non_missing\x18\x01 \x01(\x01\x12\x13\n\x0bnum_missing\x18\x02 \x01(\x01\x12\x16\n\x0e\x61vg_num_values\x18\x03 \x01(\x01\x12\x16\n\x0etot_num_values\x18\x04 \x01(\x01\"\xc3\x01\n\x0f\x43ustomStatistic\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x03num\x18\x02 \x01(\x01H\x00\x12\r\n\x03str\x18\x03 \x01(\tH\x00\x12\x36\n\thistogram\x18\x04 \x01(\x0b\x32!.facetFeatureStatistics.HistogramH\x00\x12?\n\x0erank_histogram\x18\x05 \x01(\x0b\x32%.facetFeatureStatistics.RankHistogramH\x00\x42\x05\n\x03valJ\x04\x08\x06\x10\x07\"\xb9\x02\n\x11NumericStatistics\x12>\n\x0c\x63ommon_stats\x18\x01 \x01(\x0b\x32(.facetFeatureStatistics.CommonStatistics\x12\x0c\n\x04mean\x18\x02 \x01(\x01\x12\x0f\n\x07std_dev\x18\x03 \x01(\x01\x12\x11\n\tnum_zeros\x18\x04 \x01(\x03\x12\x0b\n\x03min\x18\x05 \x01(\x01\x12\x0e\n\x06median\x18\x06 \x01(\x01\x12\x0b\n\x03max\x18\x07 \x01(\x01\x12\x35\n\nhistograms\x18\x08 \x03(\x0b\x32!.facetFeatureStatistics.Histogram\x12Q\n\x16weighted_numeric_stats\x18\t \x01(\x0b\x32\x31.facetFeatureStatistics.WeightedNumericStatistics\"\xa0\x03\n\x10StringStatistics\x12>\n\x0c\x63ommon_stats\x18\x01 \x01(\x0b\x32(.facetFeatureStatistics.CommonStatistics\x12\x0e\n\x06unique\x18\x02 \x01(\x03\x12I\n\ntop_values\x18\x03 \x03(\x0b\x32\x35.facetFeatureStatistics.StringStatistics.FreqAndValue\x12\x12\n\navg_length\x18\x04 \x01(\x02\x12=\n\x0erank_histogram\x18\x05 \x01(\x0b\x32%.facetFeatureStatistics.RankHistogram\x12O\n\x15weighted_string_stats\x18\x06 \x01(\x0b\x32\x30.facetFeatureStatistics.WeightedStringStatistics\x1aM\n\x0c\x46reqAndValue\x12\x1b\n\x0f\x64\x65precated_freq\x18\x01 \x01(\x03\x42\x02\x18\x01\x12\r\n\x05value\x18\x02 \x01(\t\x12\x11\n\tfrequency\x18\x03 \x01(\x01\"\x81\x01\n\x19WeightedNumericStatistics\x12\x0c\n\x04mean\x18\x01 \x01(\x01\x12\x0f\n\x07std_dev\x18\x02 \x01(\x01\x12\x0e\n\x06median\x18\x03 \x01(\x01\x12\x35\n\nhistograms\x18\x04 \x03(\x0b\x32!.facetFeatureStatistics.Histogram\"\xa4\x01\n\x18WeightedStringStatistics\x12I\n\ntop_values\x18\x01 \x03(\x0b\x32\x35.facetFeatureStatistics.StringStatistics.FreqAndValue\x12=\n\x0erank_histogram\x18\x02 \x01(\x0b\x32%.facetFeatureStatistics.RankHistogram\"\xac\x01\n\x0f\x42ytesStatistics\x12>\n\x0c\x63ommon_stats\x18\x01 \x01(\x0b\x32(.facetFeatureStatistics.CommonStatistics\x12\x0e\n\x06unique\x18\x02 \x01(\x03\x12\x15\n\ravg_num_bytes\x18\x03 \x01(\x02\x12\x15\n\rmin_num_bytes\x18\x04 \x01(\x02\x12\x15\n\rmax_num_bytes\x18\x05 \x01(\x02J\x04\x08\x06\x10\x07\"R\n\x10StructStatistics\x12>\n\x0c\x63ommon_stats\x18\x01 \x01(\x0b\x32(.facetFeatureStatistics.CommonStatistics\"\x88\x03\n\x10\x43ommonStatistics\x12\x17\n\x0fnum_non_missing\x18\x01 \x01(\x03\x12\x13\n\x0bnum_missing\x18\x02 \x01(\x03\x12\x16\n\x0emin_num_values\x18\x03 \x01(\x03\x12\x16\n\x0emax_num_values\x18\x04 \x01(\x03\x12\x16\n\x0e\x61vg_num_values\x18\x05 \x01(\x02\x12\x16\n\x0etot_num_values\x18\x08 \x01(\x03\x12?\n\x14num_values_histogram\x18\x06 \x01(\x0b\x32!.facetFeatureStatistics.Histogram\x12O\n\x15weighted_common_stats\x18\x07 \x01(\x0b\x32\x30.facetFeatureStatistics.WeightedCommonStatistics\x12H\n\x1d\x66\x65\x61ture_list_length_histogram\x18\t \x01(\x0b\x32!.facetFeatureStatistics.HistogramJ\x04\x08\n\x10\x0bJ\x04\x08\x0b\x10\x0c\"\xce\x02\n\tHistogram\x12\x0f\n\x07num_nan\x18\x01 \x01(\x03\x12\x15\n\rnum_undefined\x18\x02 \x01(\x03\x12\x39\n\x07\x62uckets\x18\x03 \x03(\x0b\x32(.facetFeatureStatistics.Histogram.Bucket\x12=\n\x04type\x18\x04 \x01(\x0e\x32/.facetFeatureStatistics.Histogram.HistogramType\x12\x0c\n\x04name\x18\x05 \x01(\t\x1a\x63\n\x06\x42ucket\x12\x11\n\tlow_value\x18\x01 \x01(\x01\x12\x12\n\nhigh_value\x18\x02 \x01(\x01\x12\x1c\n\x10\x64\x65precated_count\x18\x03 \x01(\x03\x42\x02\x18\x01\x12\x14\n\x0csample_count\x18\x04 \x01(\x01\",\n\rHistogramType\x12\x0c\n\x08STANDARD\x10\x00\x12\r\n\tQUANTILES\x10\x01\"\xce\x01\n\rRankHistogram\x12=\n\x07\x62uckets\x18\x01 \x03(\x0b\x32,.facetFeatureStatistics.RankHistogram.Bucket\x12\x0c\n\x04name\x18\x02 \x01(\t\x1ap\n\x06\x42ucket\x12\x10\n\x08low_rank\x18\x01 \x01(\x03\x12\x11\n\thigh_rank\x18\x02 \x01(\x03\x12\x1c\n\x10\x64\x65precated_count\x18\x03 \x01(\x03\x42\x02\x18\x01\x12\r\n\x05label\x18\x04 \x01(\t\x12\x14\n\x0csample_count\x18\x05 \x01(\x01')



  _DATASETFEATURESTATISTICSLIST = DESCRIPTOR.message_types_by_name['DatasetFeatureStatisticsList']
  _PATH = DESCRIPTOR.message_types_by_name['Path']
  _DATASETFEATURESTATISTICS = DESCRIPTOR.message_types_by_name['DatasetFeatureStatistics']
  _FEATURENAMESTATISTICS = DESCRIPTOR.message_types_by_name['FeatureNameStatistics']
  _WEIGHTEDCOMMONSTATISTICS = DESCRIPTOR.message_types_by_name['WeightedCommonStatistics']
  _CUSTOMSTATISTIC = DESCRIPTOR.message_types_by_name['CustomStatistic']
  _NUMERICSTATISTICS = DESCRIPTOR.message_types_by_name['NumericStatistics']
  _STRINGSTATISTICS = DESCRIPTOR.message_types_by_name['StringStatistics']
  _STRINGSTATISTICS_FREQANDVALUE = _STRINGSTATISTICS.nested_types_by_name['FreqAndValue']
  _WEIGHTEDNUMERICSTATISTICS = DESCRIPTOR.message_types_by_name['WeightedNumericStatistics']
  _WEIGHTEDSTRINGSTATISTICS = DESCRIPTOR.message_types_by_name['WeightedStringStatistics']
  _BYTESSTATISTICS = DESCRIPTOR.message_types_by_name['BytesStatistics']
  _STRUCTSTATISTICS = DESCRIPTOR.message_types_by_name['StructStatistics']
  _COMMONSTATISTICS = DESCRIPTOR.message_types_by_name['CommonStatistics']
  _HISTOGRAM = DESCRIPTOR.message_types_by_name['Histogram']
  _HISTOGRAM_BUCKET = _HISTOGRAM.nested_types_by_name['Bucket']
  _RANKHISTOGRAM = DESCRIPTOR.message_types_by_name['RankHistogram']
  _RANKHISTOGRAM_BUCKET = _RANKHISTOGRAM.nested_types_by_name['Bucket']
  _FEATURENAMESTATISTICS_TYPE = _FEATURENAMESTATISTICS.enum_types_by_name['Type']
  _HISTOGRAM_HISTOGRAMTYPE = _HISTOGRAM.enum_types_by_name['HistogramType']
  DatasetFeatureStatisticsList = _reflection.GeneratedProtocolMessageType('DatasetFeatureStatisticsList', (_message.Message,), {
    'DESCRIPTOR' : _DATASETFEATURESTATISTICSLIST,
    '__module__' : 'facet_feature_statistics_pb2'
    # @@protoc_insertion_point(class_scope:facetFeatureStatistics.DatasetFeatureStatisticsList)
    })
  _sym_db.RegisterMessage(DatasetFeatureStatisticsList)

  Path = _reflection.GeneratedProtocolMessageType('Path', (_message.Message,), {
    'DESCRIPTOR' : _PATH,
    '__module__' : 'facet_feature_statistics_pb2'
    # @@protoc_insertion_point(class_scope:facetFeatureStatistics.Path)
    })
  _sym_db.RegisterMessage(Path)

  DatasetFeatureStatistics = _reflection.GeneratedProtocolMessageType('DatasetFeatureStatistics', (_message.Message,), {
    'DESCRIPTOR' : _DATASETFEATURESTATISTICS,
    '__module__' : 'facet_feature_statistics_pb2'
    # @@protoc_insertion_point(class_scope:facetFeatureStatistics.DatasetFeatureStatistics)
    })
  _sym_db.RegisterMessage(DatasetFeatureStatistics)

  FeatureNameStatistics = _reflection.GeneratedProtocolMessageType('FeatureNameStatistics', (_message.Message,), {
    'DESCRIPTOR' : _FEATURENAMESTATISTICS,
    '__module__' : 'facet_feature_statistics_pb2'
    # @@protoc_insertion_point(class_scope:facetFeatureStatistics.FeatureNameStatistics)
    })
  _sym_db.RegisterMessage(FeatureNameStatistics)

  WeightedCommonStatistics = _reflection.GeneratedProtocolMessageType('WeightedCommonStatistics', (_message.Message,), {
    'DESCRIPTOR' : _WEIGHTEDCOMMONSTATISTICS,
    '__module__' : 'facet_feature_statistics_pb2'
    # @@protoc_insertion_point(class_scope:facetFeatureStatistics.WeightedCommonStatistics)
    })
  _sym_db.RegisterMessage(WeightedCommonStatistics)

  CustomStatistic = _reflection.GeneratedProtocolMessageType('CustomStatistic', (_message.Message,), {
    'DESCRIPTOR' : _CUSTOMSTATISTIC,
    '__module__' : 'facet_feature_statistics_pb2'
    # @@protoc_insertion_point(class_scope:facetFeatureStatistics.CustomStatistic)
    })
  _sym_db.RegisterMessage(CustomStatistic)

  NumericStatistics = _reflection.GeneratedProtocolMessageType('NumericStatistics', (_message.Message,), {
    'DESCRIPTOR' : _NUMERICSTATISTICS,
    '__module__' : 'facet_feature_statistics_pb2'
    # @@protoc_insertion_point(class_scope:facetFeatureStatistics.NumericStatistics)
    })
  _sym_db.RegisterMessage(NumericStatistics)

  StringStatistics = _reflection.GeneratedProtocolMessageType('StringStatistics', (_message.Message,), {

    'FreqAndValue' : _reflection.GeneratedProtocolMessageType('FreqAndValue', (_message.Message,), {
      'DESCRIPTOR' : _STRINGSTATISTICS_FREQANDVALUE,
      '__module__' : 'facet_feature_statistics_pb2'
      # @@protoc_insertion_point(class_scope:facetFeatureStatistics.StringStatistics.FreqAndValue)
      })
    ,
    'DESCRIPTOR' : _STRINGSTATISTICS,
    '__module__' : 'facet_feature_statistics_pb2'
    # @@protoc_insertion_point(class_scope:facetFeatureStatistics.StringStatistics)
    })
  _sym_db.RegisterMessage(StringStatistics)
  _sym_db.RegisterMessage(StringStatistics.FreqAndValue)

  WeightedNumericStatistics = _reflection.GeneratedProtocolMessageType('WeightedNumericStatistics', (_message.Message,), {
    'DESCRIPTOR' : _WEIGHTEDNUMERICSTATISTICS,
    '__module__' : 'facet_feature_statistics_pb2'
    # @@protoc_insertion_point(class_scope:facetFeatureStatistics.WeightedNumericStatistics)
    })
  _sym_db.RegisterMessage(WeightedNumericStatistics)

  WeightedStringStatistics = _reflection.GeneratedProtocolMessageType('WeightedStringStatistics', (_message.Message,), {
    'DESCRIPTOR' : _WEIGHTEDSTRINGSTATISTICS,
    '__module__' : 'facet_feature_statistics_pb2'
    # @@protoc_insertion_point(class_scope:facetFeatureStatistics.WeightedStringStatistics)
    })
  _sym_db.RegisterMessage(WeightedStringStatistics)

  BytesStatistics = _reflection.GeneratedProtocolMessageType('BytesStatistics', (_message.Message,), {
    'DESCRIPTOR' : _BYTESSTATISTICS,
    '__module__' : 'facet_feature_statistics_pb2'
    # @@protoc_insertion_point(class_scope:facetFeatureStatistics.BytesStatistics)
    })
  _sym_db.RegisterMessage(BytesStatistics)

  StructStatistics = _reflection.GeneratedProtocolMessageType('StructStatistics', (_message.Message,), {
    'DESCRIPTOR' : _STRUCTSTATISTICS,
    '__module__' : 'facet_feature_statistics_pb2'
    # @@protoc_insertion_point(class_scope:facetFeatureStatistics.StructStatistics)
    })
  _sym_db.RegisterMessage(StructStatistics)

  CommonStatistics = _reflection.GeneratedProtocolMessageType('CommonStatistics', (_message.Message,), {
    'DESCRIPTOR' : _COMMONSTATISTICS,
    '__module__' : 'facet_feature_statistics_pb2'
    # @@protoc_insertion_point(class_scope:facetFeatureStatistics.CommonStatistics)
    })
  _sym_db.RegisterMessage(CommonStatistics)

  Histogram = _reflection.GeneratedProtocolMessageType('Histogram', (_message.Message,), {

    'Bucket' : _reflection.GeneratedProtocolMessageType('Bucket', (_message.Message,), {
      'DESCRIPTOR' : _HISTOGRAM_BUCKET,
      '__module__' : 'facet_feature_statistics_pb2'
      # @@protoc_insertion_point(class_scope:facetFeatureStatistics.Histogram.Bucket)
      })
    ,
    'DESCRIPTOR' : _HISTOGRAM,
    '__module__' : 'facet_feature_statistics_pb2'
    # @@protoc_insertion_point(class_scope:facetFeatureStatistics.Histogram)
    })
  _sym_db.RegisterMessage(Histogram)
  _sym_db.RegisterMessage(Histogram.Bucket)

  RankHistogram = _reflection.GeneratedProtocolMessageType('RankHistogram', (_message.Message,), {

    'Bucket' : _reflection.GeneratedProtocolMessageType('Bucket', (_message.Message,), {
      'DESCRIPTOR' : _RANKHISTOGRAM_BUCKET,
      '__module__' : 'facet_feature_statistics_pb2'
      # @@protoc_insertion_point(class_scope:facetFeatureStatistics.RankHistogram.Bucket)
      })
    ,
    'DESCRIPTOR' : _RANKHISTOGRAM,
    '__module__' : 'facet_feature_statistics_pb2'
    # @@protoc_insertion_point(class_scope:facetFeatureStatistics.RankHistogram)
    })
  _sym_db.RegisterMessage(RankHistogram)
  _sym_db.RegisterMessage(RankHistogram.Bucket)

  if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _STRINGSTATISTICS_FREQANDVALUE.fields_by_name['deprecated_freq']._options = None
    _STRINGSTATISTICS_FREQANDVALUE.fields_by_name['deprecated_freq']._serialized_options = b'\030\001'
    _HISTOGRAM_BUCKET.fields_by_name['deprecated_count']._options = None
    _HISTOGRAM_BUCKET.fields_by_name['deprecated_count']._serialized_options = b'\030\001'
    _RANKHISTOGRAM_BUCKET.fields_by_name['deprecated_count']._options = None
    _RANKHISTOGRAM_BUCKET.fields_by_name['deprecated_count']._serialized_options = b'\030\001'
    _DATASETFEATURESTATISTICSLIST._serialized_start=58
    _DATASETFEATURESTATISTICSLIST._serialized_end=156
    _PATH._serialized_start=158
    _PATH._serialized_end=178
    _DATASETFEATURESTATISTICS._serialized_start=181
    _DATASETFEATURESTATISTICS._serialized_end=339
    _FEATURENAMESTATISTICS._serialized_start=342
    _FEATURENAMESTATISTICS._serialized_end=900
    _FEATURENAMESTATISTICS_TYPE._serialized_start=818
    _FEATURENAMESTATISTICS_TYPE._serialized_end=879
    _WEIGHTEDCOMMONSTATISTICS._serialized_start=902
    _WEIGHTEDCOMMONSTATISTICS._serialized_end=1022
    _CUSTOMSTATISTIC._serialized_start=1025
    _CUSTOMSTATISTIC._serialized_end=1220
    _NUMERICSTATISTICS._serialized_start=1223
    _NUMERICSTATISTICS._serialized_end=1536
    _STRINGSTATISTICS._serialized_start=1539
    _STRINGSTATISTICS._serialized_end=1955
    _STRINGSTATISTICS_FREQANDVALUE._serialized_start=1878
    _STRINGSTATISTICS_FREQANDVALUE._serialized_end=1955
    _WEIGHTEDNUMERICSTATISTICS._serialized_start=1958
    _WEIGHTEDNUMERICSTATISTICS._serialized_end=2087
    _WEIGHTEDSTRINGSTATISTICS._serialized_start=2090
    _WEIGHTEDSTRINGSTATISTICS._serialized_end=2254
    _BYTESSTATISTICS._serialized_start=2257
    _BYTESSTATISTICS._serialized_end=2429
    _STRUCTSTATISTICS._serialized_start=2431
    _STRUCTSTATISTICS._serialized_end=2513
    _COMMONSTATISTICS._serialized_start=2516
    _COMMONSTATISTICS._serialized_end=2908
    _HISTOGRAM._serialized_start=2911
    _HISTOGRAM._serialized_end=3245
    _HISTOGRAM_BUCKET._serialized_start=3100
    _HISTOGRAM_BUCKET._serialized_end=3199
    _HISTOGRAM_HISTOGRAMTYPE._serialized_start=3201
    _HISTOGRAM_HISTOGRAMTYPE._serialized_end=3245
    _RANKHISTOGRAM._serialized_start=3248
    _RANKHISTOGRAM._serialized_end=3454
    _RANKHISTOGRAM_BUCKET._serialized_start=3342
    _RANKHISTOGRAM_BUCKET._serialized_end=3454
  # @@protoc_insertion_point(module_scope)




"""Utility functions for Azure authentication."""

from __future__ import annotations

import time
import logging
from datetime import datetime, timedelta
from typing import Any, Optional
from urllib.parse import parse_qs, urlparse

logger = logging.getLogger(__name__)


def is_token_expired(token_expiry: Optional[datetime], buffer_seconds: int = 300) -> bool:
    """Check if a token is expired or will expire within the buffer time.

    Args:
        token_expiry: Token expiration datetime
        buffer_seconds: Buffer time in seconds before actual expiry

    Returns:
        True if token is expired or will expire soon
    """
    if not token_expiry:
        return True

    buffer_time = timedelta(seconds=buffer_seconds)
    return datetime.utcnow() + buffer_time >= token_expiry


def parse_connection_string(connection_string: str) -> dict[str, Any]:
    """Parse a PostgreSQL connection string and extract components.

    Args:
        connection_string: PostgreSQL connection string

    Returns:
        Dictionary with connection components
    """
    # Handle azure-postgres:// scheme by converting to postgresql:// for parsing
    original_scheme = connection_string.split("://")[0] if "://" in connection_string else "unknown"
    if connection_string.startswith("azure-postgres://"):
        connection_string = connection_string.replace("azure-postgres://", "postgresql://", 1)

    parsed = urlparse(connection_string)

    # Parse query parameters
    query_params = {}
    if parsed.query:
        query_params = {k: v[0] if len(v) == 1 else v for k, v in parse_qs(parsed.query).items()}

    return {
        "scheme": original_scheme,  # Use original scheme (azure-postgres or postgresql)
        "username": parsed.username,
        "password": parsed.password,
        "hostname": parsed.hostname,
        "port": parsed.port or 5432,
        "database": parsed.path.lstrip("/") if parsed.path else "",
        "params": query_params,
    }


def build_connection_string(
    username: str,
    password: str,
    hostname: str,
    port: int = 5432,
    database: str = "postgres",
    **params: Any,
) -> str:
    """Build a PostgreSQL connection string.

    Args:
        username: Database username
        password: Database password
        hostname: Database hostname
        port: Database port
        database: Database name
        **params: Additional connection parameters

    Returns:
        PostgreSQL connection string
    """
    param_string = ""
    if params:
        param_string = "?" + "&".join(f"{k}={v}" for k, v in params.items())

    return f"postgresql://{username}:{password}@{hostname}:{port}/{database}{param_string}"


def sanitize_connection_string_for_logging(connection_string: str) -> str:
    """Sanitize connection string for safe logging by removing sensitive information.

    Args:
        connection_string: Original connection string

    Returns:
        Sanitized connection string with password masked
    """
    try:
        parsed = parse_connection_string(connection_string)

        # Mask password
        masked_password = "***" if parsed.get("password") else None

        # Rebuild without sensitive info
        safe_parts = []
        safe_parts.append(f"scheme={parsed.get('scheme', 'unknown')}")
        safe_parts.append(f"host={parsed.get('hostname', 'unknown')}")
        safe_parts.append(f"port={parsed.get('port', 5432)}")
        safe_parts.append(f"database={parsed.get('database', 'unknown')}")
        safe_parts.append(f"user={parsed.get('username', 'unknown')}")

        if masked_password:
            safe_parts.append("password=***")

        return " | ".join(safe_parts)

    except Exception as e:
        logger.warning("Failed to sanitize connection string", extra={"error": str(e)})
        return "connection_string=<unable_to_parse>"


def exponential_backoff(attempt: int, base_delay: float = 1.0, max_delay: float = 60.0) -> float:
    """Calculate exponential backoff delay.

    Args:
        attempt: Current attempt number (0-indexed)
        base_delay: Base delay in seconds
        max_delay: Maximum delay in seconds

    Returns:
        Delay in seconds
    """
    delay = base_delay * (2**attempt)
    return min(delay, max_delay)


def retry_on_exception(
    func: callable,
    max_retries: int = 3,
    base_delay: float = 1.0,
    exceptions: tuple = (Exception,),
) -> Any:
    """Retry a function with exponential backoff.

    Args:
        func: Function to retry
        max_retries: Maximum number of retries
        base_delay: Base delay between retries
        exceptions: Exception types to catch and retry on

    Returns:
        Function result

    Raises:
        Last exception if all retries failed
    """
    last_exception = None

    for attempt in range(max_retries + 1):
        try:
            return func()
        except exceptions as e:
            last_exception = e

            if attempt == max_retries:
                logger.error(
                    "Function failed after all retries",
                    extra={"attempt": attempt, "max_retries": max_retries, "error": str(e)},
                )
                raise

            delay = exponential_backoff(attempt, base_delay)
            logger.warning(
                "Function failed, retrying",
                extra={"attempt": attempt, "delay": delay, "error": str(e)},
            )

            time.sleep(delay)

    # This should never be reached
    raise last_exception or Exception("Unexpected error in retry logic")


import base64
import json
import logging
import os
import posixpath
import uuid
from typing import Any, Optional

import requests

import mlflow.tracking
from mlflow.azure.client import (
    patch_adls_file_upload,
    patch_adls_flush,
    put_adls_file_creation,
    put_block,
    put_block_list,
)
from mlflow.entities import FileInfo
from mlflow.environment_variables import (
    MLFLOW_ASYNC_TRACE_LOGGING_RETRY_TIMEOUT,
    MLFLOW_MULTIPART_DOWNLOAD_CHUNK_SIZE,
    MLFLOW_MULTIPART_UPLOAD_CHUNK_SIZE,
    MLFLOW_MULTIPART_UPLOAD_MINIMUM_FILE_SIZE,
)
from mlflow.exceptions import (
    MlflowException,
    MlflowTraceDataCorrupted,
    MlflowTraceDataNotFound,
)
from mlflow.protos.databricks_artifacts_pb2 import (
    ArtifactCredentialType,
    CompleteMultipartUpload,
    CreateMultipartUpload,
    DatabricksMlflowArtifactsService,
    GetPresignedUploadPartUrl,
    PartEtag,
)
from mlflow.protos.databricks_pb2 import (
    INTERNAL_ERROR,
    INVALID_PARAMETER_VALUE,
)
from mlflow.protos.service_pb2 import MlflowService
from mlflow.store.artifact.artifact_repo import write_local_temp_trace_data_file
from mlflow.store.artifact.cloud_artifact_repo import (
    CloudArtifactRepository,
    _complete_futures,
    _compute_num_chunks,
    _validate_chunk_size_aws,
)
from mlflow.store.artifact.databricks_artifact_repo_resources import (
    _CredentialType,
    _LoggedModel,
    _Resource,
    _Run,
    _Trace,
)
from mlflow.tracing.constant import TRACE_REQUEST_ID_PREFIX
from mlflow.utils import chunk_list
from mlflow.utils.databricks_utils import get_databricks_host_creds
from mlflow.utils.file_utils import (
    download_file_using_http_uri,
    read_chunk,
)
from mlflow.utils.proto_json_utils import message_to_json
from mlflow.utils.request_utils import cloud_storage_http_request
from mlflow.utils.rest_utils import (
    _REST_API_PATH_PREFIX,
    augmented_raise_for_status,
    call_endpoint,
    extract_api_info_for_service,
)
from mlflow.utils.uri import (
    extract_and_normalize_path,
    get_databricks_profile_uri_from_artifact_uri,
    is_databricks_acled_artifacts_uri,
    is_valid_dbfs_uri,
    remove_databricks_profile_info_from_artifact_uri,
)

_logger = logging.getLogger(__name__)
_MAX_CREDENTIALS_REQUEST_SIZE = 2000  # Max number of artifact paths in a single credentials request
_SERVICE_AND_METHOD_TO_INFO = {
    service: extract_api_info_for_service(service, _REST_API_PATH_PREFIX)
    for service in [MlflowService, DatabricksMlflowArtifactsService]
}


class DatabricksArtifactRepository(CloudArtifactRepository):
    """
    Performs storage operations on artifacts in the access-controlled
    `dbfs:/databricks/mlflow-tracking` location.

    Signed access URIs for S3 / Azure Blob Storage are fetched from the MLflow service and used to
    read and write files from/to this location.

    The artifact_uri is expected to be in one of the following forms:
    - dbfs:/databricks/mlflow-tracking/<EXP_ID>/<RUN_ID>/
    - databricks/mlflow-tracking/<EXP_ID>/logged_models/<MODEL_ID>/artifacts/<path>
    """

    def __init__(self, artifact_uri: str, tracking_uri: Optional[str] = None) -> None:
        if not is_valid_dbfs_uri(artifact_uri):
            raise MlflowException(
                message="DBFS URI must be of the form dbfs:/<path> or "
                + "dbfs://profile@databricks/<path>",
                error_code=INVALID_PARAMETER_VALUE,
            )
        if not is_databricks_acled_artifacts_uri(artifact_uri):
            raise MlflowException(
                message=(
                    "Artifact URI incorrect. Expected path prefix to be"
                    " databricks/mlflow-tracking/path/to/artifact/.."
                ),
                error_code=INVALID_PARAMETER_VALUE,
            )
        # The dbfs:/ path ultimately used for artifact operations should not contain the
        # Databricks profile info, so strip it before setting ``artifact_uri``.
        super().__init__(
            remove_databricks_profile_info_from_artifact_uri(artifact_uri), tracking_uri
        )

        self.databricks_profile_uri = (
            get_databricks_profile_uri_from_artifact_uri(artifact_uri)
            or mlflow.tracking.get_tracking_uri()
        )
        self.resource = self._extract_resource(self.artifact_uri)

    def _extract_resource(self, artifact_uri) -> _Resource:
        """
        The artifact_uri is expected to be in one of the following formats:
        - dbfs:/databricks/mlflow-tracking/<EXP_ID>/<RUN_ID>/artifacts/<path>
        - dbfs:/databricks/mlflow-tracking/<EXP_ID>/logged_models/<MODEL_ID>/artifacts/<path>
        - databricks/mlflow-tracking/<EXP_ID>/<RUN_ID>/artifacts/<path>
        - databricks/mlflow-tracking/<EXP_ID>/logged_models/<MODEL_ID>/artifacts/<path>

        Returns:
            A `_Resource` object representing the MLflow resource associated with the specified
            artifact URI.
        """
        artifact_path = extract_and_normalize_path(artifact_uri)
        parts = artifact_path.split("/")

        if parts[3] == "logged_models":
            return _LoggedModel(
                id_=parts[4], artifact_uri=artifact_uri, call_endpoint=self._call_endpoint
            )

        if parts[3].startswith(TRACE_REQUEST_ID_PREFIX):
            return _Trace(
                id_=parts[3], artifact_uri=artifact_uri, call_endpoint=self._call_endpoint
            )

        return _Run(id_=parts[3], artifact_uri=artifact_uri, call_endpoint=self._call_endpoint)

    @staticmethod
    def _extract_run_id(artifact_uri: str) -> Optional[str]:
        """
        Extracts the run ID from the run artifact URI.
        """
        artifact_path = extract_and_normalize_path(artifact_uri)
        parts = artifact_path.split("/")
        if len(parts) < 4:
            return None

        if parts[3] == "logged_models" or parts[3].startswith(TRACE_REQUEST_ID_PREFIX):
            return None

        return parts[3]

    def _call_endpoint(
        self, service, api, json_body=None, path_params=None, retry_timeout_seconds=None
    ):
        """
        Calls the specified REST endpoint with the specified JSON body and path parameters.

        Args:
            service: The service to call.
            api: The API to call.
            json_body: The JSON body of the request.
            path_params: The path parameters to substitute into the endpoint URI.
            retry_timeout_seconds: The timeout in seconds for retrying failed requests.

        Returns:
            The response from the REST endpoint.
        """
        db_creds = get_databricks_host_creds(self.databricks_profile_uri)
        endpoint, method = _SERVICE_AND_METHOD_TO_INFO[service][api]
        if path_params:
            endpoint = endpoint.format(**path_params)
        response_proto = api.Response()

        return call_endpoint(
            host_creds=db_creds,
            endpoint=endpoint,
            method=method,
            json_body=json_body,
            response_proto=response_proto,
            retry_timeout_seconds=retry_timeout_seconds,
        )

    def _get_credential_infos(self, cred_type: _CredentialType, paths: list[str]):
        """
        Issue one or more requests for artifact credentials, providing read or write
        access to the specified resource relative artifact `paths` within the MLflow
        resource specified by `self.resource.id`. The type of access credentials, read or write,
        is specified by `request_message_class`.

        Args:
            cred_type: Specifies the type of access credentials, read or write.
            paths: The specified relative artifact paths within the MLflow resource.

        Returns:
            A list of `ArtifactCredentialInfo` objects providing read access to the specified
            relative artifact `paths` within the MLflow resource specified by `resource`.
        """
        credential_infos = []

        for paths_chunk in chunk_list(paths, _MAX_CREDENTIALS_REQUEST_SIZE):
            page_token = None
            while True:
                cred_infos, next_page_token = self.resource.get_credentials(
                    cred_type=cred_type,
                    paths=paths_chunk,
                    page_token=page_token,
                )
                credential_infos += cred_infos
                page_token = next_page_token
                if not page_token or len(cred_infos) == 0:
                    break

        return credential_infos

    def _get_write_credential_infos(self, remote_file_paths):
        """
        A list of `ArtifactCredentialInfo` objects providing write access to the specified
        relative artifact `paths` within the MLflow resource specified by `self.resource.id`.
        """
        relative_remote_paths = [
            posixpath.join(self.resource.relative_path, p or "") for p in remote_file_paths
        ]
        return self._get_credential_infos(_CredentialType.WRITE, relative_remote_paths)

    def download_trace_data(self) -> dict[str, Any]:
        [cred], _ = self.resource.get_credentials(cred_type=_CredentialType.READ)
        signed_uri = cred.signed_uri
        headers = self._extract_headers_from_credentials(cred.headers)
        with cloud_storage_http_request("get", signed_uri, headers=headers) as resp:
            try:
                augmented_raise_for_status(resp)
            except requests.HTTPError as e:
                if e.response.status_code == 404:
                    raise MlflowTraceDataNotFound(request_id=self.resource.id) from e
                raise

            try:
                return json.loads(resp.content)
            except json.JSONDecodeError as e:
                raise MlflowTraceDataCorrupted(request_id=self.resource.id) from e

    def upload_trace_data(self, trace_data: str) -> None:
        [cred], _ = self.resource.get_credentials(
            cred_type=_CredentialType.WRITE,
            timeout=MLFLOW_ASYNC_TRACE_LOGGING_RETRY_TIMEOUT.get(),
        )
        with write_local_temp_trace_data_file(trace_data) as temp_file:
            if cred.type == ArtifactCredentialType.AZURE_ADLS_GEN2_SAS_URI:
                self._azure_adls_gen2_upload_file(
                    credentials=cred,
                    local_file=temp_file,
                    artifact_file_path=None,
                    get_credentials=lambda artifact_paths: [
                        self._get_upload_trace_data_cred_info()
                    ],
                )
            elif cred.type == ArtifactCredentialType.AZURE_SAS_URI:
                self._azure_upload_file(
                    credentials=cred,
                    local_file=temp_file,
                    artifact_file_path=None,
                    get_credentials=lambda artifact_paths: [
                        self._get_upload_trace_data_cred_info()
                    ],
                )
            elif (
                cred.type == ArtifactCredentialType.AWS_PRESIGNED_URL
                or cred.type == ArtifactCredentialType.GCP_SIGNED_URL
            ):
                self._signed_url_upload_file(cred, temp_file)

    def _get_read_credential_infos(self, remote_file_paths):
        """
        Returns:
            A list of `ArtifactCredentialInfo` objects providing read access to the specified
            relative artifact `paths` within the MLflow resource specified.
        """
        if type(remote_file_paths) == str:
            remote_file_paths = [remote_file_paths]
        if type(remote_file_paths) != list:
            raise MlflowException(
                f"Expected `paths` to be a list of strings. Got {type(remote_file_paths)}"
            )
        relative_remote_paths = [
            posixpath.join(self.resource.relative_path, p) for p in remote_file_paths
        ]
        return self._get_credential_infos(_CredentialType.READ, relative_remote_paths)

    def _extract_headers_from_credentials(self, headers):
        """
        Returns:
            A python dictionary of http headers converted from the protobuf credentials.
        """
        return {header.name: header.value for header in headers}

    def _azure_upload_chunk(
        self,
        credentials,
        headers,
        local_file,
        artifact_file_path,
        start_byte,
        size,
        get_credentials,
    ):
        """
        Uploads a chunk of a file to a given Azure storage location.

        Args:
            credentials: The credentials for the upload.
            headers: The headers for the upload.
            local_file: The local file to upload.
            artifact_file_path: The path to the artifact file.
            start_byte: The starting byte of the chunk.
            size: The size of the chunk.
            get_credentials: The function to call to get new credentials.
        """
        # Base64-encode a UUID, producing a UTF8-encoded bytestring. Then, decode
        # the bytestring for compliance with Azure Blob Storage API requests
        block_id = base64.b64encode(uuid.uuid4().hex.encode()).decode("utf-8")
        chunk = read_chunk(local_file, size, start_byte)
        try:
            put_block(credentials.signed_uri, block_id, chunk, headers=headers)
        except requests.HTTPError as e:
            if e.response.status_code in [401, 403]:
                _logger.info(
                    "Failed to authorize request, possibly due to credential expiration."
                    " Refreshing credentials and trying again..."
                )
                credential_info = get_credentials([artifact_file_path])[0]
                put_block(credential_info.signed_uri, block_id, chunk, headers=headers)
            else:
                raise e
        return block_id

    def _azure_upload_file(self, credentials, local_file, artifact_file_path, get_credentials):
        """
        Uploads a file to a given Azure storage location.
        The function uses a file chunking generator with 100 MB being the size limit for each chunk.
        This limit is imposed by the stage_block API in azure-storage-blob.
        In the case the file size is large and the upload takes longer than the validity of the
        given credentials, a new set of credentials are generated and the operation continues. This
        is the reason for the first nested try-except block
        Finally, since the prevailing credentials could expire in the time between the last
        stage_block and the commit, a second try-except block refreshes credentials if needed.

        Args:
            credentials: The credentials for the upload.
            local_file: The local file to upload.
            artifact_file_path: The path to the artifact file.
            get_credentials: The function to call to get new credentials.
        """
        try:
            headers = self._extract_headers_from_credentials(credentials.headers)
            futures = {}
            num_chunks = _compute_num_chunks(local_file, MLFLOW_MULTIPART_UPLOAD_CHUNK_SIZE.get())
            for index in range(num_chunks):
                start_byte = index * MLFLOW_MULTIPART_UPLOAD_CHUNK_SIZE.get()
                future = self.chunk_thread_pool.submit(
                    self._azure_upload_chunk,
                    credentials=credentials,
                    headers=headers,
                    local_file=local_file,
                    artifact_file_path=artifact_file_path,
                    start_byte=start_byte,
                    size=MLFLOW_MULTIPART_UPLOAD_CHUNK_SIZE.get(),
                    get_credentials=get_credentials,
                )
                futures[future] = index

            results, errors = _complete_futures(futures, local_file)
            if errors:
                raise MlflowException(
                    f"Failed to upload at least one part of {local_file}. Errors: {errors}"
                )
            # Sort results by the chunk index
            uploading_block_list = [results[index] for index in sorted(results)]

            try:
                put_block_list(credentials.signed_uri, uploading_block_list, headers=headers)
            except requests.HTTPError as e:
                if e.response.status_code in [401, 403]:
                    _logger.info(
                        "Failed to authorize request, possibly due to credential expiration."
                        " Refreshing credentials and trying again..."
                    )
                    credential_info = get_credentials([artifact_file_path])[0]
                    put_block_list(
                        credential_info.signed_uri, uploading_block_list, headers=headers
                    )
                else:
                    raise e
        except Exception as err:
            raise MlflowException(err)

    def _retryable_adls_function(self, func, artifact_file_path, get_credentials, **kwargs):
        """
        Calls the passed function, retrying if the credentials have expired.

        Args:
            func: The function to call.
            artifact_file_path: The artifact file path.
            get_credentials: The function to call to get new credentials.
            **kwargs: The keyword arguments to pass to the function.
        """
        # Attempt to call the passed function.  Retry if the credentials have expired
        try:
            func(**kwargs)
        except requests.HTTPError as e:
            if e.response.status_code in [403]:
                _logger.info(
                    "Failed to authorize ADLS operation, possibly due "
                    "to credential expiration. Refreshing credentials and trying again..."
                )
                new_credentials = get_credentials([artifact_file_path])[0]
                kwargs["sas_url"] = new_credentials.signed_uri
                func(**kwargs)
            else:
                raise e

    def _azure_adls_gen2_upload_file(
        self, credentials, local_file, artifact_file_path, get_credentials
    ):
        """
        Uploads a file to a given Azure storage location using the ADLS gen2 API.

        Args:
            credentials: The credentials for the upload.
            local_file: The local file to upload.
            artifact_file_path: The path to the artifact file.
            get_credentials: The function to call to get new credentials.
        """
        try:
            headers = self._extract_headers_from_credentials(credentials.headers)

            # try to create the file
            self._retryable_adls_function(
                func=put_adls_file_creation,
                artifact_file_path=artifact_file_path,
                get_credentials=get_credentials,
                sas_url=credentials.signed_uri,
                headers=headers,
            )

            # next try to append the file
            futures = {}
            file_size = os.path.getsize(local_file)
            num_chunks = _compute_num_chunks(local_file, MLFLOW_MULTIPART_UPLOAD_CHUNK_SIZE.get())
            use_single_part_upload = num_chunks == 1
            for index in range(num_chunks):
                start_byte = index * MLFLOW_MULTIPART_UPLOAD_CHUNK_SIZE.get()
                future = self.chunk_thread_pool.submit(
                    self._retryable_adls_function,
                    func=patch_adls_file_upload,
                    artifact_file_path=artifact_file_path,
                    get_credentials=get_credentials,
                    sas_url=credentials.signed_uri,
                    local_file=local_file,
                    start_byte=start_byte,
                    size=MLFLOW_MULTIPART_UPLOAD_CHUNK_SIZE.get(),
                    position=start_byte,
                    headers=headers,
                    is_single=use_single_part_upload,
                )
                futures[future] = index

            _, errors = _complete_futures(futures, local_file)
            if errors:
                raise MlflowException(
                    f"Failed to upload at least one part of {artifact_file_path}. Errors: {errors}"
                )

            # finally try to flush the file
            if not use_single_part_upload:
                self._retryable_adls_function(
                    func=patch_adls_flush,
                    artifact_file_path=artifact_file_path,
                    get_credentials=get_credentials,
                    sas_url=credentials.signed_uri,
                    position=file_size,
                    headers=headers,
                )
        except Exception as err:
            raise MlflowException(err)

    def _signed_url_upload_file(self, credentials, local_file):
        try:
            headers = self._extract_headers_from_credentials(credentials.headers)
            signed_write_uri = credentials.signed_uri
            # Putting an empty file in a request by reading file bytes gives 501 error.
            if os.stat(local_file).st_size == 0:
                with cloud_storage_http_request(
                    "put", signed_write_uri, data="", headers=headers
                ) as response:
                    augmented_raise_for_status(response)
            else:
                with open(local_file, "rb") as file:
                    with cloud_storage_http_request(
                        "put", signed_write_uri, data=file, headers=headers
                    ) as response:
                        augmented_raise_for_status(response)
        except Exception as err:
            raise MlflowException(err)

    def _upload_to_cloud(self, cloud_credential_info, src_file_path, artifact_file_path):
        """
        Upload a local file to the cloud. Note that in this artifact repository, files are uploaded
        to resource relative artifact file paths in the artifact repository.

        Args:
            cloud_credential_info: ArtifactCredentialInfo object with presigned URL for the file.
            src_file_path: Local source file path for the upload.
            artifact_file_path: Path in the artifact repository, relative to the resource root path,
                where the artifact will be logged.

        """
        if cloud_credential_info.type == ArtifactCredentialType.AZURE_SAS_URI:
            self._azure_upload_file(
                cloud_credential_info,
                src_file_path,
                artifact_file_path,
                get_credentials=self._get_write_credential_infos,
            )
        elif cloud_credential_info.type == ArtifactCredentialType.AZURE_ADLS_GEN2_SAS_URI:
            self._azure_adls_gen2_upload_file(
                cloud_credential_info,
                src_file_path,
                artifact_file_path,
                self._get_write_credential_infos,
            )
        elif cloud_credential_info.type == ArtifactCredentialType.AWS_PRESIGNED_URL:
            if os.path.getsize(src_file_path) > MLFLOW_MULTIPART_UPLOAD_MINIMUM_FILE_SIZE.get():
                _validate_chunk_size_aws(MLFLOW_MULTIPART_UPLOAD_CHUNK_SIZE.get())
                self._multipart_upload(src_file_path, artifact_file_path)
            else:
                self._signed_url_upload_file(cloud_credential_info, src_file_path)
        elif cloud_credential_info.type == ArtifactCredentialType.GCP_SIGNED_URL:
            self._signed_url_upload_file(cloud_credential_info, src_file_path)
        else:
            raise MlflowException(
                message="Cloud provider not supported.", error_code=INTERNAL_ERROR
            )

    def _download_from_cloud(self, remote_file_path, local_path):
        """
        Download a file from the input `remote_file_path` and save it to `local_path`.

        Args:
            remote_file_path: Path relative to the resource root path to file in remote artifact
                repository.
            local_path: Local path to download file to.

        """
        read_credentials = self._get_read_credential_infos(remote_file_path)
        # Read credentials for only one file were requested. So we expected only one value in
        # the response.
        assert len(read_credentials) == 1
        cloud_credential_info = read_credentials[0]

        if cloud_credential_info.type not in [
            ArtifactCredentialType.AZURE_SAS_URI,
            ArtifactCredentialType.AZURE_ADLS_GEN2_SAS_URI,
            ArtifactCredentialType.AWS_PRESIGNED_URL,
            ArtifactCredentialType.GCP_SIGNED_URL,
        ]:
            raise MlflowException(
                message="Cloud provider not supported.", error_code=INTERNAL_ERROR
            )
        try:
            download_file_using_http_uri(
                cloud_credential_info.signed_uri,
                local_path,
                MLFLOW_MULTIPART_DOWNLOAD_CHUNK_SIZE.get(),
                self._extract_headers_from_credentials(cloud_credential_info.headers),
            )
        except Exception as err:
            raise MlflowException(err)

    def _create_multipart_upload(self, run_id, path, num_parts):
        return self._call_endpoint(
            DatabricksMlflowArtifactsService,
            CreateMultipartUpload,
            message_to_json(CreateMultipartUpload(run_id=run_id, path=path, num_parts=num_parts)),
        )

    def _get_presigned_upload_part_url(self, run_id, path, upload_id, part_number):
        return self._call_endpoint(
            DatabricksMlflowArtifactsService,
            GetPresignedUploadPartUrl,
            message_to_json(
                GetPresignedUploadPartUrl(
                    run_id=run_id, path=path, upload_id=upload_id, part_number=part_number
                )
            ),
        )

    def _upload_part(self, cred_info, data):
        headers = self._extract_headers_from_credentials(cred_info.headers)
        with cloud_storage_http_request(
            "put",
            cred_info.signed_uri,
            data=data,
            headers=headers,
        ) as response:
            augmented_raise_for_status(response)
            return response.headers["ETag"]

    def _upload_part_retry(self, cred_info, upload_id, part_number, local_file, start_byte, size):
        data = read_chunk(local_file, size, start_byte)
        try:
            return self._upload_part(cred_info, data)
        except requests.HTTPError as e:
            if e.response.status_code not in (401, 403):
                raise e
            _logger.info(
                "Failed to authorize request, possibly due to credential expiration."
                " Refreshing credentials and trying again..."
            )
            resp = self._get_presigned_upload_part_url(
                cred_info.run_id, cred_info.path, upload_id, part_number
            )
            return self._upload_part(resp.upload_credential_info, data)

    def _upload_parts(self, local_file, create_mpu_resp):
        futures = {}
        for index, cred_info in enumerate(create_mpu_resp.upload_credential_infos):
            part_number = index + 1
            start_byte = index * MLFLOW_MULTIPART_UPLOAD_CHUNK_SIZE.get()
            future = self.chunk_thread_pool.submit(
                self._upload_part_retry,
                cred_info=cred_info,
                upload_id=create_mpu_resp.upload_id,
                part_number=part_number,
                local_file=local_file,
                start_byte=start_byte,
                size=MLFLOW_MULTIPART_UPLOAD_CHUNK_SIZE.get(),
            )
            futures[future] = part_number

        results, errors = _complete_futures(futures, local_file)
        if errors:
            raise MlflowException(
                f"Failed to upload at least one part of {local_file}. Errors: {errors}"
            )

        return [
            PartEtag(part_number=part_number, etag=results[part_number])
            for part_number in sorted(results)
        ]

    def _complete_multipart_upload(self, run_id, path, upload_id, part_etags):
        return self._call_endpoint(
            DatabricksMlflowArtifactsService,
            CompleteMultipartUpload,
            message_to_json(
                CompleteMultipartUpload(
                    run_id=run_id,
                    path=path,
                    upload_id=upload_id,
                    part_etags=part_etags,
                )
            ),
        )

    def _abort_multipart_upload(self, cred_info):
        headers = self._extract_headers_from_credentials(cred_info.headers)
        with cloud_storage_http_request(
            "delete", cred_info.signed_uri, headers=headers
        ) as response:
            augmented_raise_for_status(response)
            return response

    def _multipart_upload(self, local_file, artifact_file_path):
        run_relative_artifact_path = posixpath.join(
            self.resource.relative_path, artifact_file_path or ""
        )
        num_parts = _compute_num_chunks(local_file, MLFLOW_MULTIPART_UPLOAD_CHUNK_SIZE.get())
        create_mpu_resp = self._create_multipart_upload(
            self.resource.id, run_relative_artifact_path, num_parts
        )
        try:
            part_etags = self._upload_parts(local_file, create_mpu_resp)
            self._complete_multipart_upload(
                self.resource.id,
                run_relative_artifact_path,
                create_mpu_resp.upload_id,
                part_etags,
            )
        except Exception as e:
            _logger.warning(
                "Encountered an unexpected error during multipart upload: %s, aborting", e
            )
            self._abort_multipart_upload(create_mpu_resp.abort_credential_info)
            raise e

    def log_artifact(self, local_file, artifact_path=None):
        src_file_name = os.path.basename(local_file)
        artifact_file_path = posixpath.join(artifact_path or "", src_file_name)
        write_credential_info = self._get_write_credential_infos([artifact_file_path])[0]
        self._upload_to_cloud(
            cloud_credential_info=write_credential_info,
            src_file_path=local_file,
            artifact_file_path=artifact_file_path,
        )

    def list_artifacts(self, path: Optional[str] = None) -> list[FileInfo]:
        return self.resource.list_artifacts(path)

    def delete_artifacts(self, artifact_path=None):
        raise MlflowException("Not implemented yet")



import logging
from typing import TYPE_CHECKING, Optional, Union

from opentelemetry.sdk.trace import Span

from mlflow.tracing.utils import set_span_chat_messages, set_span_chat_tools
from mlflow.types.chat import (
    ChatMessage,
    ChatTool,
    Function,
    TextContentPart,
    ToolCall,
)

if TYPE_CHECKING:
    from autogen_core import FunctionCall
    from autogen_core.models import LLMMessage
    from autogen_core.tools import BaseTool, ToolSchema

_logger = logging.getLogger(__name__)


def log_tools(span: Span, tools: list[Union["BaseTool", "ToolSchema"]]):
    """
    Log Autogen tool definitions into the passed in span.

    Ref: https://microsoft.github.io/autogen/stable/user-guide/core-user-guide/components/tools.html

    Args:
        span: The span to log the tools into.
        tools: A list of Autogen BaseTool.
    """
    from autogen_core.tools import BaseTool

    try:
        tools = [
            ChatTool(
                type="function",
                function=tool.schema if isinstance(tool, BaseTool) else tool,
            )
            for tool in tools
        ]
        set_span_chat_tools(span, tools)
    except Exception:
        _logger.debug(f"Failed to log tools to Span {span}.", exc_info=True)


def convert_assistant_message_to_chat_message(
    content: Union[str, list["FunctionCall"]],
) -> Optional[ChatMessage]:
    """
    Convert an Autogen assistant message to a ChatMessage.
    The content of assistant message is a str or a list of tool calls.

    Args:
        content: The Autogen assistant message content to convert.

    Returns:
        A ChatMessage object
    """
    from autogen_core import FunctionCall

    if isinstance(content, str):
        return ChatMessage(role="assistant", content=content)
    elif isinstance(content, list) and all(isinstance(f, FunctionCall) for f in content):
        return ChatMessage(
            role="assistant",
            tool_calls=[
                ToolCall(
                    id=f.id, type="function", function=Function(name=f.name, arguments=f.arguments)
                )
                for f in content
            ],
        )
    else:
        _logger.debug(f"Unsupported message type: {type(content)}. Skipping conversion.")


def log_chat_messages(span: Span, messages: list["LLMMessage"]):
    """
    Log Autogen chat messages into the passed in span.

    Args:
        span: The span to log the tools into.
        messages: A list of Autogen chat messages.
    """
    from autogen_core.models import (
        AssistantMessage,
        FunctionExecutionResultMessage,
        SystemMessage,
        UserMessage,
    )

    chat_messages = []
    for message in messages:
        if isinstance(message, SystemMessage):
            chat_messages.append(
                ChatMessage(
                    role="system",
                    content=message.content,
                )
            )
        elif isinstance(message, UserMessage):
            content = message.content
            if isinstance(content, list):
                parts = []
                for part in content:
                    if isinstance(part, str):
                        parts.append(
                            TextContentPart(
                                type="text",
                                text=part,
                            )
                        )
                    else:
                        # The content type of UserMessage is text or image
                        parts.append(part.to_openai_format())
                content = parts
            chat_messages.append(
                ChatMessage(
                    role="user",
                    content=content,
                )
            )
        elif isinstance(message, AssistantMessage):
            content = message.content
            if chat_message := convert_assistant_message_to_chat_message(content):
                chat_messages.append(chat_message)
        elif isinstance(message, FunctionExecutionResultMessage):
            chat_messages.append(
                ChatMessage(
                    role="user",
                    content=message.model_dump(),
                )
            )
        else:
            _logger.debug(f"Unsupported message type: {type(message)}. Skipping logging.")

    try:
        set_span_chat_messages(span, chat_messages)
    except Exception:
        _logger.debug(f"Failed to log chat messages to Span {span}.", exc_info=True)



from typing import Optional

from mlflow.entities.logged_model_parameter import LoggedModelParameter as ModelParam
from mlflow.entities.metric import Metric
from mlflow.entities.model_registry._model_registry_entity import _ModelRegistryEntity
from mlflow.entities.model_registry.model_version_deployment_job_state import (
    ModelVersionDeploymentJobState,
)
from mlflow.entities.model_registry.model_version_status import ModelVersionStatus
from mlflow.entities.model_registry.model_version_tag import ModelVersionTag
from mlflow.protos.model_registry_pb2 import ModelVersion as ProtoModelVersion
from mlflow.protos.model_registry_pb2 import ModelVersionTag as ProtoModelVersionTag


class ModelVersion(_ModelRegistryEntity):
    """
    MLflow entity for Model Version.
    """

    def __init__(
        self,
        name: str,
        version: str,
        creation_timestamp: int,
        last_updated_timestamp: Optional[int] = None,
        description: Optional[str] = None,
        user_id: Optional[str] = None,
        current_stage: Optional[str] = None,
        source: Optional[str] = None,
        run_id: Optional[str] = None,
        status: str = ModelVersionStatus.to_string(ModelVersionStatus.READY),
        status_message: Optional[str] = None,
        tags: Optional[list[ModelVersionTag]] = None,
        run_link: Optional[str] = None,
        aliases: Optional[list[str]] = None,
        # TODO: Make model_id a required field
        # (currently optional to minimize breakages during prototype development)
        model_id: Optional[str] = None,
        params: Optional[list[ModelParam]] = None,
        metrics: Optional[list[Metric]] = None,
        deployment_job_state: Optional[ModelVersionDeploymentJobState] = None,
    ):
        super().__init__()
        self._name: str = name
        self._version: str = version
        self._creation_time: int = creation_timestamp
        self._last_updated_timestamp: Optional[int] = last_updated_timestamp
        self._description: Optional[str] = description
        self._user_id: Optional[str] = user_id
        self._current_stage: Optional[str] = current_stage
        self._source: Optional[str] = source
        self._run_id: Optional[str] = run_id
        self._run_link: Optional[str] = run_link
        self._status: str = status
        self._status_message: Optional[str] = status_message
        self._tags: dict[str, str] = {tag.key: tag.value for tag in (tags or [])}
        self._aliases: list[str] = aliases or []
        self._model_id: Optional[str] = model_id
        self._params: Optional[list[ModelParam]] = params
        self._metrics: Optional[list[Metric]] = metrics
        self._deployment_job_state: Optional[ModelVersionDeploymentJobState] = deployment_job_state

    @property
    def name(self) -> str:
        """String. Unique name within Model Registry."""
        return self._name

    @name.setter
    def name(self, new_name: str):
        self._name = new_name

    @property
    def version(self) -> str:
        """Version"""
        return self._version

    @property
    def creation_timestamp(self) -> int:
        """Integer. Model version creation timestamp (milliseconds since the Unix epoch)."""
        return self._creation_time

    @property
    def last_updated_timestamp(self) -> Optional[int]:
        """Integer. Timestamp of last update for this model version (milliseconds since the Unix
        epoch).
        """
        return self._last_updated_timestamp

    @last_updated_timestamp.setter
    def last_updated_timestamp(self, updated_timestamp: int):
        self._last_updated_timestamp = updated_timestamp

    @property
    def description(self) -> Optional[str]:
        """String. Description"""
        return self._description

    @description.setter
    def description(self, description: str):
        self._description = description

    @property
    def user_id(self) -> Optional[str]:
        """String. User ID that created this model version."""
        return self._user_id

    @property
    def current_stage(self) -> Optional[str]:
        """String. Current stage of this model version."""
        return self._current_stage

    @current_stage.setter
    def current_stage(self, stage: str):
        self._current_stage = stage

    @property
    def source(self) -> Optional[str]:
        """String. Source path for the model."""
        return self._source

    @property
    def run_id(self) -> Optional[str]:
        """String. MLflow run ID that generated this model."""
        return self._run_id

    @property
    def run_link(self) -> Optional[str]:
        """String. MLflow run link referring to the exact run that generated this model version."""
        return self._run_link

    @property
    def status(self) -> str:
        """String. Current Model Registry status for this model."""
        return self._status

    @property
    def status_message(self) -> Optional[str]:
        """String. Descriptive message for error status conditions."""
        return self._status_message

    @property
    def tags(self) -> dict[str, str]:
        """Dictionary of tag key (string) -> tag value for the current model version."""
        return self._tags

    @property
    def aliases(self) -> list[str]:
        """List of aliases (string) for the current model version."""
        return self._aliases

    @aliases.setter
    def aliases(self, aliases: list[str]):
        self._aliases = aliases

    @property
    def model_id(self) -> Optional[str]:
        """String. ID of the model associated with this version."""
        return self._model_id

    @property
    def params(self) -> Optional[list[ModelParam]]:
        """List of parameters associated with this model version."""
        return self._params

    @property
    def metrics(self) -> Optional[list[Metric]]:
        """List of metrics associated with this model version."""
        return self._metrics

    @property
    def deployment_job_state(self) -> Optional[ModelVersionDeploymentJobState]:
        """Deployment job state for the current model version."""
        return self._deployment_job_state

    @classmethod
    def _properties(cls) -> list[str]:
        # aggregate with base class properties since cls.__dict__ does not do it automatically
        return sorted(cls._get_properties_helper())

    def _add_tag(self, tag: ModelVersionTag):
        self._tags[tag.key] = tag.value

    # proto mappers
    @classmethod
    def from_proto(cls, proto) -> "ModelVersion":
        # input: mlflow.protos.model_registry_pb2.ModelVersion
        # returns: ModelVersion entity
        model_version = cls(
            proto.name,
            proto.version,
            proto.creation_timestamp,
            proto.last_updated_timestamp,
            proto.description if proto.HasField("description") else None,
            proto.user_id,
            proto.current_stage,
            proto.source,
            proto.run_id if proto.HasField("run_id") else None,
            ModelVersionStatus.to_string(proto.status),
            proto.status_message if proto.HasField("status_message") else None,
            run_link=proto.run_link,
            aliases=proto.aliases,
            deployment_job_state=ModelVersionDeploymentJobState.from_proto(
                proto.deployment_job_state
            ),
        )
        for tag in proto.tags:
            model_version._add_tag(ModelVersionTag.from_proto(tag))
        # TODO: Include params, metrics, and model ID in proto
        return model_version

    def to_proto(self):
        # input: ModelVersion entity
        # returns mlflow.protos.model_registry_pb2.ModelVersion
        model_version = ProtoModelVersion()
        model_version.name = self.name
        model_version.version = str(self.version)
        model_version.creation_timestamp = self.creation_timestamp
        if self.last_updated_timestamp is not None:
            model_version.last_updated_timestamp = self.last_updated_timestamp
        if self.description is not None:
            model_version.description = self.description
        if self.user_id is not None:
            model_version.user_id = self.user_id
        if self.current_stage is not None:
            model_version.current_stage = self.current_stage
        if self.source is not None:
            model_version.source = str(self.source)
        if self.run_id is not None:
            model_version.run_id = str(self.run_id)
        if self.run_link is not None:
            model_version.run_link = str(self.run_link)
        if self.status is not None:
            model_version.status = ModelVersionStatus.from_string(self.status)
        if self.status_message:
            model_version.status_message = self.status_message
        model_version.tags.extend(
            [ProtoModelVersionTag(key=key, value=value) for key, value in self._tags.items()]
        )
        model_version.aliases.extend(self.aliases)
        if self.deployment_job_state is not None:
            ModelVersionDeploymentJobState.to_proto(self.deployment_job_state)
        # TODO: Include params, metrics, and model ID in proto
        return model_version



import logging
import os
from typing import Optional

_logger = logging.getLogger(__name__)

MLServerMLflowRuntime = "mlserver_mlflow.MLflowRuntime"
MLServerDefaultModelName = "mlflow-model"


def get_cmd(
    model_uri: str,
    port: Optional[int] = None,
    host: Optional[str] = None,
    timeout: Optional[int] = None,
    nworkers: Optional[int] = None,
    model_name: Optional[str] = None,
    model_version: Optional[str] = None,
) -> tuple[str, dict[str, str]]:
    cmd = f"mlserver start {model_uri}"

    cmd_env = os.environ.copy()

    if port:
        cmd_env["MLSERVER_HTTP_PORT"] = str(port)

    if host:
        cmd_env["MLSERVER_HOST"] = host

    if timeout:
        _logger.warning("Timeout is not yet supported in MLServer.")

    if nworkers:
        cmd_env["MLSERVER_PARALLEL_WORKERS"] = str(nworkers)

    # give precedence to user env var input
    cmd_env["MLSERVER_MODEL_NAME"] = (
        cmd_env.get("MLSERVER_MODEL_NAME") or model_name or MLServerDefaultModelName
    )
    if model_version and not cmd_env.get("MLSERVER_MODEL_VERSION"):
        cmd_env["MLSERVER_MODEL_VERSION"] = model_version

    cmd_env["MLSERVER_MODEL_IMPLEMENTATION"] = MLServerMLflowRuntime
    cmd_env["MLSERVER_MODEL_URI"] = model_uri

    return cmd, cmd_env



import json
from typing import Any, Union

from pydantic import BaseModel

from mlflow.exceptions import MlflowException
from mlflow.types.chat import (
    ChatMessage,
    ChatTool,
    Function,
    FunctionToolDefinition,
    ImageContentPart,
    ImageUrl,
    TextContentPart,
    ToolCall,
)
from mlflow.utils import IS_PYDANTIC_V2_OR_NEWER


def convert_message_to_mlflow_chat(message: Union[BaseModel, dict[str, Any]]) -> ChatMessage:
    """
    Convert Anthropic message object into MLflow's standard format (OpenAI compatible).

    Ref: https://docs.anthropic.com/en/api/messages#body-messages

    Args:
        message: Anthropic message object or a dictionary representing the message.

    Returns:
        ChatMessage: MLflow's standard chat message object.
    """
    if isinstance(message, dict):
        content = message.get("content")
        role = message.get("role")
    elif isinstance(message, BaseModel):
        content = message.content
        role = message.role
    else:
        raise MlflowException.invalid_parameter_value(
            f"Message must be either a dict or a Message object, but got: {type(message)}."
        )

    if isinstance(content, str):
        return ChatMessage(role=role, content=content)

    elif isinstance(content, list):
        contents = []
        tool_calls = []
        tool_call_id = None
        for content_block in content:
            if isinstance(content_block, BaseModel):
                if IS_PYDANTIC_V2_OR_NEWER:
                    content_block = content_block.model_dump()
                else:
                    content_block = content_block.dict()

            content_type = content_block.get("type")
            if content_type == "tool_use":
                # Anthropic response contains tool calls in the content block
                # Ref: https://docs.anthropic.com/en/docs/build-with-claude/tool-use#example-api-response-with-a-tool-use-content-block
                tool_calls.append(
                    ToolCall(
                        id=content_block["id"],
                        function=Function(
                            name=content_block["name"], arguments=json.dumps(content_block["input"])
                        ),
                        type="function",
                    )
                )
            elif content_type == "tool_result":
                # In Anthropic, the result of tool execution is returned as a special content type
                # "tool_result" with "user" role, which corresponds to the "tool" role in OpenAI.
                role = "tool"
                tool_call_id = content_block["tool_use_id"]
                if result_content := content_block.get("content"):
                    contents.append(_parse_content(result_content))
                else:
                    contents.append(TextContentPart(text="", type="text"))
            else:
                contents.append(_parse_content(content_block))

        message = ChatMessage(role=role, content=contents)
        # Only set tool_calls field when it is present
        if tool_calls:
            message.tool_calls = tool_calls
        if tool_call_id:
            message.tool_call_id = tool_call_id
        return message

    else:
        raise MlflowException.invalid_parameter_value(
            f"Invalid content type. Must be either a string or a list, but got: {type(content)}."
        )


def _parse_content(content: Union[str, dict[str, Any]]) -> Union[TextContentPart, ImageContentPart]:
    if isinstance(content, str):
        return TextContentPart(text=content, type="text")

    content_type = content.get("type")
    if content_type == "text":
        return TextContentPart(text=content["text"], type="text")
    elif content_type == "image":
        source = content["source"]
        return ImageContentPart(
            image_url=ImageUrl(
                url=f"data:{source['media_type']};{source['type']},{source['data']}"
            ),
            type="image_url",
        )
    # Claude 3.7 added new "thinking" content block, which is essentially a text block as of now.
    # TODO: We should consider adding a new ContentPart type if more providers support this.
    # https://docs.anthropic.com/en/docs/build-with-claude/extended-thinking
    elif content_type == "thinking":
        return TextContentPart(text=content["thinking"], type="text")
    else:
        raise MlflowException.invalid_parameter_value(
            f"Unknown content type: {content_type['type']}. Please make sure the message "
            "is a valid Anthropic message object. If it is a valid type, contact to the "
            "MLflow maintainer via https://github.com/mlflow/mlflow/issues/new/choose for "
            "requesting support for a new message type."
        )


def convert_tool_to_mlflow_chat_tool(tool: dict[str, Any]) -> ChatTool:
    """
    Convert Anthropic tool definition into MLflow's standard format (OpenAI compatible).

    Ref: https://docs.anthropic.com/en/docs/build-with-claude/tool-use

    Args:
        tool: A dictionary represents a single tool definition in the input request.

    Returns:
        ChatTool: MLflow's standard tool definition object.
    """
    return ChatTool(
        type="function",
        function=FunctionToolDefinition(
            name=tool.get("name"),
            description=tool.get("description"),
            parameters=tool.get("input_schema"),
        ),
    )



"""
Databricks Agent Datasets Python SDK. For more details see Databricks Agent Evaluation:
 <https://docs.databricks.com/en/generative-ai/agent-evaluation/index.html>

The API docs can be found here:
<https://api-docs.databricks.com/python/databricks-agents/latest/databricks_agent_eval.html#datasets>
"""

from typing import Optional, Union

from mlflow.genai.datasets.evaluation_dataset import EvaluationDataset

_ERROR_MSG = (
    "The `databricks-agents` package is required to use `mlflow.genai.datasets`. "
    "Please install it with `pip install databricks-agents`."
)


def create_dataset(
    uc_table_name: str, experiment_id: Optional[Union[str, list[str]]] = None
) -> "EvaluationDataset":
    """Create a dataset with the given name and associate it with the given experiment.

    Args:
        uc_table_name: The UC table name of the dataset.
        experiment_id: The ID of the experiment to associate the dataset with. If not provided,
            the current experiment is inferred from the environment.
    """
    try:
        from databricks.agents.datasets import create_dataset
    except ImportError as e:
        raise ImportError(_ERROR_MSG) from e
    return EvaluationDataset(create_dataset(uc_table_name, experiment_id))


def delete_dataset(uc_table_name: str) -> None:
    """Delete the dataset with the given name.

    Args:
        uc_table_name: The UC table name of the dataset.
    """
    try:
        from databricks.agents.datasets import delete_dataset
    except ImportError:
        raise ImportError(_ERROR_MSG) from None
    return delete_dataset(uc_table_name)


def get_dataset(uc_table_name: str) -> "EvaluationDataset":
    """Get the dataset with the given name.

    Args:
        uc_table_name: The UC table name of the dataset.
    """
    try:
        from databricks.agents.datasets import get_dataset
    except ImportError as e:
        raise ImportError(_ERROR_MSG) from e
    return EvaluationDataset(get_dataset(uc_table_name))


__all__ = [
    "create_dataset",
    "delete_dataset",
    "get_dataset",
    "EvaluationDataset",
]



"""Update run status constraint with killed

Revision ID: cfd24bdc0731
Revises: 89d4b8295536
Create Date: 2019-10-11 15:55:10.853449

"""

import alembic
from alembic import op
from packaging.version import Version
from sqlalchemy import CheckConstraint, Enum

from mlflow.entities import RunStatus, ViewType
from mlflow.entities.lifecycle_stage import LifecycleStage
from mlflow.store.tracking.dbmodels.models import SourceTypes, SqlRun

# revision identifiers, used by Alembic.
revision = "cfd24bdc0731"
down_revision = "2b4d017a5e9b"
branch_labels = None
depends_on = None

old_run_statuses = [
    RunStatus.to_string(RunStatus.SCHEDULED),
    RunStatus.to_string(RunStatus.FAILED),
    RunStatus.to_string(RunStatus.FINISHED),
    RunStatus.to_string(RunStatus.RUNNING),
]

new_run_statuses = [*old_run_statuses, RunStatus.to_string(RunStatus.KILLED)]

# Certain SQL backends (e.g., SQLite) do not preserve CHECK constraints during migrations.
# For these backends, CHECK constraints must be specified as table arguments. Here, we define
# the collection of CHECK constraints that should be preserved when performing the migration.
# The "status" constraint is excluded from this set because it is explicitly modified
# within the migration's `upgrade()` routine.
check_constraint_table_args = [
    CheckConstraint(SqlRun.source_type.in_(SourceTypes), name="source_type"),
    CheckConstraint(
        SqlRun.lifecycle_stage.in_(LifecycleStage.view_type_to_stages(ViewType.ALL)),
        name="runs_lifecycle_stage",
    ),
]


def upgrade():
    # In alembic >= 1.7.0, `table_args` is unnecessary since CHECK constraints are preserved
    # during migrations.
    table_args = (
        [] if Version(alembic.__version__) >= Version("1.7.0") else check_constraint_table_args
    )
    with op.batch_alter_table("runs", table_args=table_args) as batch_op:
        # Transform the "status" column to an `Enum` and define a new check constraint. Specify
        # `native_enum=False` to create a check constraint rather than a
        # database-backend-dependent enum (see https://docs.sqlalchemy.org/en/13/core/
        # type_basics.html#sqlalchemy.types.Enum.params.native_enum)
        batch_op.alter_column(
            "status",
            type_=Enum(
                *new_run_statuses,
                create_constraint=True,
                native_enum=False,
            ),
            existing_type=Enum(
                *old_run_statuses,
                create_constraint=True,
                native_enum=False,
                name="status",
            ),
        )


def downgrade():
    # Omit downgrade logic for now - we don't currently provide users a command/API for
    # reverting a database migration, instead recommending that they take a database backup
    # before running the migration.
    pass



from mlflow.utils.pydantic_utils import IS_PYDANTIC_V2_OR_NEWER, model_validator

if not IS_PYDANTIC_V2_OR_NEWER:
    raise ImportError(
        "mlflow.types.responses is not supported in Pydantic v1. "
        "Please upgrade to Pydantic v2 or newer."
    )
import json
from typing import Any, Optional, Union

from pydantic import ConfigDict

from mlflow.types.agent import ChatContext
from mlflow.types.chat import BaseModel
from mlflow.types.responses_helpers import (
    BaseRequestPayload,
    Message,
    OutputItem,
    Response,
    ResponseCompletedEvent,
    ResponseErrorEvent,
    ResponseOutputItemDoneEvent,
    ResponseTextAnnotationDeltaEvent,
    ResponseTextDeltaEvent,
)

__all__ = [
    "ResponsesAgentRequest",
    "ResponsesAgentResponse",
    "ResponsesAgentStreamEvent",
]

from mlflow.types.schema import Schema
from mlflow.types.type_hints import _infer_schema_from_type_hint
from mlflow.utils.autologging_utils.logging_and_warnings import (
    MlflowEventsAndWarningsBehaviorGlobally,
)


class ResponsesAgentRequest(BaseRequestPayload):
    """Request object for ResponsesAgent.

    Args:
        input: List of simple `role` and `content` messages or output items. See examples at
            https://mlflow.org/docs/latest/llms/responses-agent-intro/#testing-out-your-agent and
            https://mlflow.org/docs/latest/llms/responses-agent-intro/#creating-agent-output.
        custom_inputs (Dict[str, Any]): An optional param to provide arbitrary additional context
            to the model. The dictionary values must be JSON-serializable.
            **Optional** defaults to ``None``
        context (:py:class:`mlflow.types.agent.ChatContext`): The context to be used in the chat
            endpoint. Includes conversation_id and user_id. **Optional** defaults to ``None``
    """

    input: list[Union[Message, OutputItem]]
    custom_inputs: Optional[dict[str, Any]] = None
    context: Optional[ChatContext] = None


class ResponsesAgentResponse(Response):
    """Response object for ResponsesAgent.

    Args:
        output: List of output items. See examples at
            https://mlflow.org/docs/latest/llms/responses-agent-intro/#creating-agent-output.
        reasoning: Reasoning parameters
        usage: Usage information
        custom_outputs (Dict[str, Any]): An optional param to provide arbitrary additional context
            from the model. The dictionary values must be JSON-serializable. **Optional**, defaults
            to ``None``
    """

    custom_outputs: Optional[dict[str, Any]] = None


class ResponsesAgentStreamEvent(BaseModel):
    """Stream event for ResponsesAgent.
    See examples at https://mlflow.org/docs/latest/llms/responses-agent-intro/#streaming-agent-output

    Args:
        type (str): Type of the stream event
        custom_outputs (Dict[str, Any]): An optional param to provide arbitrary additional context
            from the model. The dictionary values must be JSON-serializable. **Optional**, defaults
            to ``None``
    """

    model_config = ConfigDict(extra="allow")
    type: str
    custom_outputs: Optional[dict[str, Any]] = None

    @model_validator(mode="after")
    def check_type(self) -> "ResponsesAgentStreamEvent":
        type = self.type
        if type == "response.output_item.done":
            ResponseOutputItemDoneEvent(**self.model_dump_compat())
        elif type == "response.output_text.delta":
            ResponseTextDeltaEvent(**self.model_dump_compat())
        elif type == "response.output_text.annotation.added":
            ResponseTextAnnotationDeltaEvent(**self.model_dump_compat())
        elif type == "error":
            ResponseErrorEvent(**self.model_dump_compat())
        elif type == "response.completed":
            ResponseCompletedEvent(**self.model_dump_compat())
        """
        unvalidated types: {
            "response.created",
            "response.in_progress",
            "response.completed",
            "response.failed",
            "response.incomplete",
            "response.content_part.added",
            "response.content_part.done",
            "response.output_text.done",
            "response.output_item.added",
            "response.refusal.delta",
            "response.refusal.done",
            "response.function_call_arguments.delta",
            "response.function_call_arguments.done",
            "response.file_search_call.in_progress",
            "response.file_search_call.searching",
            "response.file_search_call.completed",
            "response.web_search_call.in_progress",
            "response.web_search_call.searching",
            "response.web_search_call.completed",
            "response.error",
        }
        """
        return self


with MlflowEventsAndWarningsBehaviorGlobally(
    reroute_warnings=False,
    disable_event_logs=True,
    disable_warnings=True,
):
    properties = _infer_schema_from_type_hint(ResponsesAgentRequest).to_dict()[0]["properties"]
    formatted_properties = [{**prop, "name": name} for name, prop in properties.items()]
    RESPONSES_AGENT_INPUT_SCHEMA = Schema.from_json(json.dumps(formatted_properties))
    RESPONSES_AGENT_OUTPUT_SCHEMA = _infer_schema_from_type_hint(ResponsesAgentResponse)
RESPONSES_AGENT_INPUT_EXAMPLE = {"input": [{"role": "user", "content": "Hello!"}]}



import warnings
from contextlib import contextmanager
from typing import Optional, Union

import mlflow.tracking._model_registry.fluent as registry_api
from mlflow.entities.model_registry.prompt import Prompt
from mlflow.entities.model_registry.prompt_version import PromptVersion
from mlflow.prompt.registry_utils import require_prompt_registry
from mlflow.store.entities.paged_list import PagedList
from mlflow.utils.annotations import experimental


@contextmanager
def suppress_genai_migration_warning():
    """Suppress the deprecation warning when the api is called from `mlflow.genai` namespace."""
    with warnings.catch_warnings():
        warnings.filterwarnings(
            action="ignore",
            category=FutureWarning,
            message="The `mlflow.*` API is moved to the `mlflow.genai` namespace.*",
        )
        yield


@experimental(version="3.0.0")
@require_prompt_registry
def register_prompt(
    name: str,
    template: str,
    commit_message: Optional[str] = None,
    tags: Optional[dict[str, str]] = None,
) -> PromptVersion:
    """
    Register a new :py:class:`Prompt <mlflow.entities.Prompt>` in the MLflow Prompt Registry.

    A :py:class:`Prompt <mlflow.entities.Prompt>` is a pair of name and
    template text at minimum. With MLflow Prompt Registry, you can create, manage, and
    version control prompts with the MLflow's robust model tracking framework.

    If there is no registered prompt with the given name, a new prompt will be created.
    Otherwise, a new version of the existing prompt will be created.


    Args:
        name: The name of the prompt.
        template: The template text of the prompt. It can contain variables enclosed in
            double curly braces, e.g. {variable}, which will be replaced with actual values
            by the `format` method.

            .. note::

                If you want to use the prompt with a framework that uses single curly braces
                e.g. LangChain, you can use the `to_single_brace_format` method to convert the
                loaded prompt to a format that uses single curly braces.

                .. code-block:: python

                    prompt = client.load_prompt("my_prompt")
                    langchain_format = prompt.to_single_brace_format()

        commit_message: A message describing the changes made to the prompt, similar to a
            Git commit message. Optional.
        tags: A dictionary of tags associated with the **prompt version**.
            This is useful for storing version-specific information, such as the author of
            the changes. Optional.

    Returns:
        A :py:class:`Prompt <mlflow.entities.Prompt>` object that was created.

    Example:

    .. code-block:: python

        import mlflow

        # Register a new prompt
        mlflow.genai.register_prompt(
            name="my_prompt",
            template="Respond to the user's message as a {{style}} AI.",
        )

        # Load the prompt from the registry
        prompt = mlflow.genai.load_prompt("my_prompt")

        # Use the prompt in your application
        import openai

        openai_client = openai.OpenAI()
        openai_client.chat.completion.create(
            model="gpt-4o-mini",
            messages=[
                {"role": "system", "content": prompt.format(style="friendly")},
                {"role": "user", "content": "Hello, how are you?"},
            ],
        )

        # Update the prompt with a new version
        prompt = mlflow.genai.register_prompt(
            name="my_prompt",
            template="Respond to the user's message as a {{style}} AI. {{greeting}}",
            commit_message="Add a greeting to the prompt.",
            tags={"author": "Bob"},
        )
    """
    with suppress_genai_migration_warning():
        return registry_api.register_prompt(
            name=name,
            template=template,
            commit_message=commit_message,
            tags=tags,
        )


@experimental(version="3.0.0")
@require_prompt_registry
def search_prompts(
    filter_string: Optional[str] = None,
    max_results: Optional[int] = None,
) -> PagedList[Prompt]:
    with suppress_genai_migration_warning():
        return registry_api.search_prompts(filter_string=filter_string, max_results=max_results)


@experimental(version="3.0.0")
@require_prompt_registry
def load_prompt(
    name_or_uri: str, version: Optional[Union[str, int]] = None, allow_missing: bool = False
) -> PromptVersion:
    """
    Load a :py:class:`Prompt <mlflow.entities.Prompt>` from the MLflow Prompt Registry.

    The prompt can be specified by name and version, or by URI.

    Args:
        name_or_uri: The name of the prompt, or the URI in the format "prompts:/name/version".
        version: The version of the prompt (required when using name, not allowed when using URI).
        allow_missing: If True, return None instead of raising Exception if the specified prompt
            is not found.

    Example:

    .. code-block:: python

        import mlflow

        # Load a specific version of the prompt
        prompt = mlflow.genai.load_prompt("my_prompt", version=1)

        # Load a specific version of the prompt by URI
        prompt = mlflow.genai.load_prompt("prompts:/my_prompt/1")

        # Load a prompt version with an alias "production"
        prompt = mlflow.genai.load_prompt("prompts:/my_prompt@production")

    """
    with suppress_genai_migration_warning():
        return registry_api.load_prompt(
            name_or_uri=name_or_uri, version=version, allow_missing=allow_missing
        )


@experimental(version="3.0.0")
@require_prompt_registry
def set_prompt_alias(name: str, alias: str, version: int) -> None:
    """
    Set an alias for a :py:class:`Prompt <mlflow.entities.Prompt>` in the MLflow Prompt Registry.

    Args:
        name: The name of the prompt.
        alias: The alias to set for the prompt.
        version: The version of the prompt.

    Example:

    .. code-block:: python

        import mlflow

        # Set an alias for the prompt
        mlflow.genai.set_prompt_alias(name="my_prompt", version=1, alias="production")

        # Load the prompt by alias (use "@" to specify the alias)
        prompt = mlflow.genai.load_prompt("prompts:/my_prompt@production")

        # Switch the alias to a new version of the prompt
        mlflow.genai.set_prompt_alias(name="my_prompt", version=2, alias="production")

        # Delete the alias
        mlflow.genai.delete_prompt_alias(name="my_prompt", alias="production")
    """
    with suppress_genai_migration_warning():
        return registry_api.set_prompt_alias(name=name, version=version, alias=alias)


@experimental(version="3.0.0")
@require_prompt_registry
def delete_prompt_alias(name: str, alias: str) -> None:
    """
    Delete an alias for a :py:class:`Prompt <mlflow.entities.Prompt>` in the MLflow Prompt Registry.

    Args:
        name: The name of the prompt.
        alias: The alias to delete for the prompt.
    """
    with suppress_genai_migration_warning():
        return registry_api.delete_prompt_alias(name=name, alias=alias)



import logging
from typing import Any, Callable, Optional

from opentelemetry.trace import NoOpTracer

import mlflow
from mlflow.entities.span import Span, SpanType
from mlflow.entities.trace import Trace
from mlflow.genai.utils.data_validation import check_model_prediction
from mlflow.tracing.constant import TraceTagKey
from mlflow.tracing.display.display_handler import IPythonTraceDisplayHandler
from mlflow.tracking.client import MlflowClient

_logger = logging.getLogger(__name__)


def convert_predict_fn(predict_fn: Callable[..., Any], sample_input: Any) -> Callable[..., Any]:
    """
    Check the predict_fn is callable and add trace decorator if it is not already traced.
    """
    with NoOpTracerPatcher() as counter:
        check_model_prediction(predict_fn, sample_input)

    if counter.count == 0:
        predict_fn = mlflow.trace(predict_fn)

    # Wrap the prediction function to unwrap the inputs dictionary into keyword arguments.
    return lambda request: predict_fn(**request)


class NoOpTracerPatcher:
    """
    A context manager to count the number of times NoOpTracer's start_span is called.

    The check is done in the following steps so it doesn't have any side effects:
    1. Disable tracing.
    2. Patch the NoOpTracer.start_span method to count the number of times it is called.
        NoOpTracer is used when tracing is disabled.
    3. Call the predict function with the sample input.
    4. Restore the original NoOpTracer.start_span method and re-enable tracing.


    WARNING: This function is not thread-safe. We do not provide support for running
        `mlflow.genai.evaluate` in multi-threaded environments.`
    """

    def __init__(self):
        self.count = 0

    def __enter__(self):
        self.original = NoOpTracer.start_span

        def _patched_start_span(_self, *args, **kwargs):
            self.count += 1
            return self.original(_self, *args, **kwargs)

        NoOpTracer.start_span = _patched_start_span
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        NoOpTracer.start_span = self.original


def parse_inputs_to_str(inputs: Any) -> str:
    """Parse the inputs to a request string compatible with the judges API"""
    from databricks.rag_eval.utils import input_output_utils

    return input_output_utils.request_to_string(inputs)


def parse_output_to_str(output: Any) -> str:
    """Parse the output to a string compatible with the judges API"""
    from databricks.rag_eval.utils import input_output_utils

    return input_output_utils.response_to_string(output)


def extract_retrieval_context_from_trace(trace: Optional[Trace]) -> dict[str, list[Any]]:
    """
    Extract the retrieval context from the trace.
    Only consider the last retrieval span in the trace if there are multiple retrieval spans.
    If the trace does not have a retrieval span, return None.
    ⚠️ Warning: Please make sure to not throw exception. If fails, return None.
    """
    if trace is None or trace.data is None:
        return {}

    # Only consider the top-level retrieval spans
    top_level_retrieval_spans = _get_top_level_retrieval_spans(trace)
    if len(top_level_retrieval_spans) == 0:
        return {}

    retrieved = {}  # span_id -> list of documents

    for retrieval_span in top_level_retrieval_spans:
        try:
            contexts = [_parse_chunk(chunk) for chunk in retrieval_span.outputs or []]
            retrieved[retrieval_span.span_id] = [c for c in contexts if c is not None]
        except Exception as e:
            _logger.debug(
                f"Fail to get retrieval context from span: {retrieval_span}. Error: {e!r}"
            )

    return retrieved


def _get_top_level_retrieval_spans(trace: Trace) -> list[Span]:
    """
    Get the top-level retrieval spans in the trace.
    Top-level retrieval spans are retrieval spans that are not children of other retrieval spans.
    For example, given the following spans:
    - Span A (Chain)
      - Span B (Retriever)
        - Span C (Retriever)
      - Span D (Retriever)
        - Span E (LLM)
          - Span F (Retriever)
    Span B and Span D are top-level retrieval spans.
    Span C and Span F are NOT top-level because they are children of other retrieval spans.
    """
    top_level_retrieval_spans = []
    # Cache span_id -> span mapping for fast lookup
    all_spans = {span.span_id: span for span in trace.data.spans}
    for span in trace.search_spans(span_type=SpanType.RETRIEVER):
        # Check if this span is a child of another retrieval span
        parent_id = span.parent_id
        while parent_id:
            parent_span = all_spans.get(parent_id)
            if not parent_span:
                # Malformed trace
                _logger.debug(
                    f"Malformed trace: span {span} has parent span ID {parent_id}, "
                    "but the parent span is not found in the trace."
                )
                break

            if parent_span.span_type == SpanType.RETRIEVER:
                # This span is a child of another retrieval span
                break

            parent_id = parent_span.parent_id
        else:
            # If the loop completes without breaking, this is a top-level span
            top_level_retrieval_spans.append(span)

    return top_level_retrieval_spans


def _parse_chunk(chunk: Any) -> Optional[dict[str, Any]]:
    if not isinstance(chunk, dict):
        return None

    doc = {"content": chunk.get("page_content")}
    if doc_uri := chunk.get("metadata", {}).get("doc_uri"):
        doc["doc_uri"] = doc_uri
    return doc


def clean_up_extra_traces(run_id: str, start_time_ms: int):
    """
    Clean up noisy traces generated outside predict function.

    Evaluation run should only contain traces that is being evaluated or generated by the predict
    function. If not, the result will not show the correct list of traces.
    Sometimes, there are extra traces generated during the evaluation, for example, custom scorer
    code might generate traces. This function cleans up those noisy traces.

    TODO: This is not a fundamental solution. Ideally, evaluation result should be able to render
        correct result even if there are extra traces in the run.

    Args:
        run_id: The ID of the run to clean up.
        start_time_ms: The start time of the evaluation in milliseconds.
    """
    from mlflow.tracking.fluent import _get_experiment_id

    try:
        # Search for all traces generated during evaluation
        traces = mlflow.search_traces(
            run_id=run_id,
            # Not download spans for efficiency
            include_spans=False,
            # Limit to traces generated after evaluation time to ensure we will not
            # delete traces generated before evaluation.
            filter_string=f"trace.timestamp >= {start_time_ms}",
            return_type="list",
        )
        extra_trace_ids = [
            # Traces from predict function should always have the EVAL_REQUEST_ID tag
            trace.info.trace_id
            for trace in traces
            if TraceTagKey.EVAL_REQUEST_ID not in trace.info.tags
        ]
        if extra_trace_ids:
            _logger.debug(
                f"Found {len(extra_trace_ids)} extra traces generated during evaluation run. "
                "Deleting them."
            )
            MlflowClient().delete_traces(
                experiment_id=_get_experiment_id(), trace_ids=extra_trace_ids
            )
            # Avoid displaying the deleted trace in notebook cell output
            for trace_id in extra_trace_ids:
                IPythonTraceDisplayHandler.get_instance().traces_to_display.pop(trace_id, None)
        else:
            _logger.debug("No extra traces found during evaluation run.")
    except Exception as e:
        _logger.warning(
            f"Failed to clean up extra traces generated during evaluation. The "
            f"result page might not show the correct list of traces. Error: {e}"
        )



import logging
from typing import Optional, Sequence

from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export import SpanExporter

from mlflow.entities.model_registry import PromptVersion
from mlflow.entities.trace import Trace
from mlflow.environment_variables import (
    MLFLOW_ENABLE_ASYNC_TRACE_LOGGING,
)
from mlflow.tracing.client import TracingClient
from mlflow.tracing.constant import TraceTagKey
from mlflow.tracing.display import get_display_handler
from mlflow.tracing.export.async_export_queue import AsyncTraceExportQueue, Task
from mlflow.tracing.export.utils import try_link_prompts_to_trace
from mlflow.tracing.fluent import _EVAL_REQUEST_ID_TO_TRACE_ID, _set_last_active_trace_id
from mlflow.tracing.trace_manager import InMemoryTraceManager
from mlflow.tracing.utils import add_size_stats_to_trace_metadata, maybe_get_request_id
from mlflow.utils.databricks_utils import is_in_databricks_notebook
from mlflow.utils.uri import is_databricks_uri

_logger = logging.getLogger(__name__)


class MlflowV3SpanExporter(SpanExporter):
    """
    An exporter implementation that logs the traces to MLflow Tracking Server
    using the V3 trace schema and API.
    """

    def __init__(self, tracking_uri: Optional[str] = None):
        self._client = TracingClient(tracking_uri)
        self._is_async_enabled = self._should_enable_async_logging()
        if self._is_async_enabled:
            self._async_queue = AsyncTraceExportQueue()

        # Display handler is no-op when running outside of notebooks.
        self._display_handler = get_display_handler()

    def export(self, spans: Sequence[ReadableSpan]):
        """
        Export the spans to the destination.

        Args:
            spans: A sequence of OpenTelemetry ReadableSpan objects passed from
                a span processor. Only root spans for each trace should be exported.
        """
        for span in spans:
            if span._parent is not None:
                _logger.debug("Received a non-root span. Skipping export.")
                continue

            manager_trace = InMemoryTraceManager.get_instance().pop_trace(span.context.trace_id)
            if manager_trace is None:
                _logger.debug(f"Trace for span {span} not found. Skipping export.")
                continue

            trace = manager_trace.trace
            _set_last_active_trace_id(trace.info.request_id)

            # Store mapping from eval request ID to trace ID so that the evaluation
            # harness can access to the trace using mlflow.get_trace(eval_request_id)
            if eval_request_id := trace.info.tags.get(TraceTagKey.EVAL_REQUEST_ID):
                _EVAL_REQUEST_ID_TO_TRACE_ID[eval_request_id] = trace.info.trace_id

            if not maybe_get_request_id(is_evaluate=True):
                self._display_handler.display_traces([trace])

            if self._should_log_async():
                self._async_queue.put(
                    task=Task(
                        handler=self._log_trace,
                        args=(trace, manager_trace.prompts),
                        error_msg="Failed to log trace to the trace server.",
                    )
                )
            else:
                self._log_trace(trace, prompts=manager_trace.prompts)

    def _log_trace(self, trace: Trace, prompts: Sequence[PromptVersion]):
        """
        Handles exporting a trace to MLflow using the V3 API and blob storage.
        Steps:
        1. Create the trace in MLflow
        2. Upload the trace data to blob storage using the returned trace info.
        """
        try:
            if trace:
                add_size_stats_to_trace_metadata(trace)
                returned_trace_info = self._client.start_trace(trace.info)
                self._client._upload_trace_data(returned_trace_info, trace.data)
            else:
                _logger.warning("No trace or trace info provided, unable to export")
        except Exception as e:
            _logger.warning(f"Failed to send trace to MLflow backend: {e}")

        try:
            # Always run prompt linking asynchronously since (1) prompt linking API calls
            # would otherwise add latency to the export procedure and (2) prompt linking is not
            # critical for trace export (if the prompt fails to link, the user's workflow is
            # minorly affected), so we don't have to await successful linking
            try_link_prompts_to_trace(
                client=self._client,
                trace_id=trace.info.trace_id,
                prompts=prompts,
                synchronous=False,
            )
        except Exception as e:
            _logger.warning(f"Failed to link prompts to trace: {e}")

    def _should_enable_async_logging(self):
        if (
            is_in_databricks_notebook()
            # NB: Not defaulting OSS backend to async logging for now to reduce blast radius.
            or not is_databricks_uri(self._client.tracking_uri)
        ):
            # NB: We don't turn on async logging in Databricks notebook by default
            # until we are confident that the async logging is working on the
            # offline workload on Databricks, to derisk the inclusion to the
            # standard image. When it is enabled explicitly via the env var, we
            # will respect that.
            return (
                MLFLOW_ENABLE_ASYNC_TRACE_LOGGING.get()
                if MLFLOW_ENABLE_ASYNC_TRACE_LOGGING.is_set()
                else False
            )

        return MLFLOW_ENABLE_ASYNC_TRACE_LOGGING.get()

    def _should_log_async(self):
        # During evaluate, the eval harness relies on the generated trace objects,
        # so we should not log traces asynchronously.
        if maybe_get_request_id(is_evaluate=True):
            return False

        return self._is_async_enabled
